import numpy as np
import matplotlib.pyplot as plt
import psyneulink as pnl
import psyneulink.core.components.functions.transferfunctions

ci = pnl.TransferMechanism(size=2, name='COLORS INPUT')
wi = pnl.TransferMechanism(size=2, name='WORDS INPUT')
ch = pnl.TransferMechanism(size=2, function=psyneulink.core.components.functions.transferfunctions.Logistic, name='COLORS HIDDEN')
wh = pnl.TransferMechanism(size=2, function=psyneulink.core.components.functions.transferfunctions.Logistic, name='WORDS HIDDEN')
tl = pnl.TransferMechanism(size=2, function=psyneulink.core.components.functions.transferfunctions.Logistic(gain=pnl.CONTROL), name='TASK CONTROL')
rl = pnl.LCAMechanism(size=2, function=psyneulink.core.components.functions.transferfunctions.Logistic, name='RESPONSE')
cp = pnl.Process(pathway=[ci, ch, rl])
wp = pnl.Process(pathway=[wi, wh, rl])
tc = pnl.Process(pathway=[tl, ch])
tw = pnl.Process(pathway=[tl,wh])
s = pnl.System(processes=[tc, tw, cp, wp],
               controller=pnl.EVCControlMechanism(name='EVC Mechanimsm'),
               monitor_for_control=[rl])
s.show_graph()



from . import controlmechanism
from . import defaultcontrolmechanism
from . import optimizationcontrolmechanism

from .controlmechanism import *
from .defaultcontrolmechanism import *
from .optimizationcontrolmechanism import *

__all__= list(controlmechanism.__all__)
__all__.extend(defaultcontrolmechanism.__all__)
__all__.extend(defaultcontrolmechanism.__all__)
__all__.extend(optimizationcontrolmechanism.__all__)




# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.


# *************************************************  EVCAuxiliary ******************************************************

"""
Auxiliary functions for `EVCControlMechanism`.

"""

import numpy as np
import typecheck as tc
import copy
import warnings

from psyneulink.core.components.functions.function import Function_Base
from psyneulink.core.components.functions.statefulfunctions.statefulfunction import StatefulFunction
from psyneulink.core.components.functions.statefulfunctions.integratorfunctions import IntegratorFunction
from psyneulink.core.components.functions.statefulfunctions.memoryfunctions import Buffer
from psyneulink.core.components.functions.transferfunctions import Linear
from psyneulink.core.components.mechanisms.processing.integratormechanism import IntegratorMechanism
from psyneulink.core.globals.context import ContextFlags, handle_external_context
from psyneulink.core.globals.defaults import MPI_IMPLEMENTATION, defaultControlAllocation
from psyneulink.core.globals.keywords import COMBINE_OUTCOME_AND_COST_FUNCTION, COST_FUNCTION, EVC_SIMULATION, FUNCTION, FUNCTION_PARAMS, NOISE, PREDICTION_MECHANISM, RATE, \
    kwPreferenceSetName, kwProgressBarChar
from psyneulink.core.globals.parameters import Parameter
from psyneulink.core.globals.preferences.componentpreferenceset import is_pref_set, kpReportOutputPref
from psyneulink.core.globals.preferences.preferenceset import PreferenceEntry, PreferenceLevel

__all__ = [
    'AVERAGE_INPUTS', 'CONTROL_SIGNAL_GRID_SEARCH_FUNCTION', 'CONTROLLER', 'ControlSignalGridSearch',
    'EVCAuxiliaryError', 'WINDOW_SIZE',
    'kwEVCAuxFunction', 'kwEVCAuxFunctionType', 'kwValueFunction',
    'INPUT', 'INPUT_SEQUENCE', 'PredictionMechanism', 'PY_MULTIPROCESSING',
    'TIME_AVERAGE_INPUT', 'ValueFunction', 'FILTER_FUNCTION'
]

PY_MULTIPROCESSING = False

if PY_MULTIPROCESSING:
    from multiprocessing import Pool


if MPI_IMPLEMENTATION:
    from mpi4py import MPI

kwEVCAuxFunction = "EVC AUXILIARY FUNCTION"
kwEVCAuxFunctionType = "EVC AUXILIARY FUNCTION TYPE"
kwValueFunction = "EVC VALUE FUNCTION"
CONTROL_SIGNAL_GRID_SEARCH_FUNCTION = "EVC CONTROL SIGNAL GRID SEARCH FUNCTION"
CONTROLLER = 'controller'


class EVCAuxiliaryError(Exception):
    def __init__(self, error_value):
        self.error_value = error_value

    def __str__(self):
        return repr(self.error_value)

class EVCAuxiliaryFunction(Function_Base):
    """Base class for EVC auxiliary functions
    """
    componentType = kwEVCAuxFunctionType

    class Parameters(Function_Base.Parameters):
        """
            Attributes
            ----------

                filter_function
                    see `filter_function <PredictionMechanism.filter_function>`

                    :default value: None
                    :type:

                rate
                    see `rate <PredictionMechanism.rate>`

                    :default value: 1.0
                    :type: float

                window_size
                    see `window_size <PredictionMechanism.window_size>`

                    :default value: 1
                    :type: int

        """
        variable = Parameter(None, read_only=True)

    classPreferences = {
        kwPreferenceSetName: 'ValueFunctionCustomClassPreferences',
        kpReportOutputPref: PreferenceEntry(False, PreferenceLevel.INSTANCE),
       }

    @tc.typecheck
    def __init__(self,
                 function,
                 variable=None,
                 params=None,
                 owner=None,
                 prefs:is_pref_set=None,
                 context=None):

        # Assign args to params and functionParams dicts
        params = self._assign_args_to_param_dicts(params=params)
        self.aux_function = function

        super().__init__(default_variable=variable,
                         params=params,
                         owner=owner,
                         prefs=prefs,
                         context=context,
                         function=function,
                         )

class ValueFunction(EVCAuxiliaryFunction):
    """Calculate the `EVC <EVCControlMechanism_EVC>` for a given performance outcome and set of costs.

    ValueFunction takes as its arguments an outcome (a value representing the performance of a `System`)
    and list of costs (each reflecting the `cost <ControlSignal.cost>` of a `ControlSignal` of the `controller
    System.controller` of that System), and returns an `expected value of control (EVC) <EVCControlMechanism_EVC>`
    based on these (along with the outcome and combination of costs used to calculate the EVC).

    ValueFunction is the default for an EVCControlMechanism's `value_function <EVCControlMechanism.value_function>`
    attribute, and it is called by `ControlSignalGridSearch` (the EVCControlMechanism's default `function
    <EVCControlMechanism.function>`).

    The ValueFunction's default `function <ValueFunction.function>` calculates the EVC using the result of the
    `function <ObjectiveMechanism.function` of the EVCControlMechanism's `objective_mechanism
    <EVCControlMechanism.objective_mechanism>`, and two auxiliary functions specified in corresponding attributes of
    the EVCControlMechanism: `cost_function <EVCControlMechanism.cost_function>` and `combine_outcome_and_cost_function
    <EVCControlMechanism.combine_outcome_and_cost_function>`. The calculation of the EVC
    provided by ValueFunction can be modified by customizing or replacing any of these functions, or by replacing the
    ValueFunction's `function <ValueFunction.function>` itself (in the EVCControlMechanism's `value_function
    <EVCControlMechanism.value_function>` attribute). Replacement functions must use the same format (number and type
    of items) for its arguments and return values (see `note <EVCControlMechanism_Calling_and_Assigning_Functions>`).

    """

    componentName = kwValueFunction

    def __init__(self, function=None):
        function = function or self.function
        super().__init__(function=function,
                         )

    def _function(
        self,
        controller=None,
        outcome=None,
        costs=None,
        variable=None,
        context=None,
        params=None,

    ):
        """
        function (controller, outcome, costs)

        Calculate the EVC as follows:

        * call the `cost_function` for the EVCControlMechanism specified in the **controller** argument,
          to combine the list of costs specified in the **costs** argument into a single cost value;

        * call the `combine_outcome_and_cost_function` for the EVCControlMechanism specified in the **controller**
          argument, to combine the value specified in the **outcome** argument with the value returned by the
         `cost_function`;

        * return the results in a three item tuple: (EVC, outcome and cost).


        Arguments
        ---------

        controller : EVCControlMechanism
            the EVCControlMechanism for which the EVC is to be calculated;  this is required so that the controller's
            `cost_function <EVCControlMechanism.cost_function>` and `combine_outcome_and_cost_function
            <EVCControlMechanism.combine_outcome_and_cost_function>` functions can be called.

        outcome : value : default float
            should represent the outcome of performance of the `System` for which an `control_allocation` is being
            evaluated.

        costs : list or array of values : default 1d np.array of floats
            each item should be the `cost <ControlSignal.cost>` of one of the controller's `ControlSignals
            <EVCControlMechanism_ControlSignals>`.


        Returns
        -------

        (EVC, outcome, cost) : Tuple(float, float, float)



        """

        if self.is_initializing:
            return (np.array([0]), np.array([0]), np.array([0]))

        # remove this in favor of attribute or parameter?
        cost_function = controller.paramsCurrent[COST_FUNCTION]
        combine_function = controller.paramsCurrent[COMBINE_OUTCOME_AND_COST_FUNCTION]

        from psyneulink.core.components.functions.userdefinedfunction import UserDefinedFunction

        # Aggregate costs
        if isinstance(cost_function, UserDefinedFunction):
            cost = cost_function._execute(controller=controller, costs=costs, context=context)
        else:
            cost = cost_function._execute(variable=costs, context=context)

        # Combine outcome and cost to determine value
        if isinstance(combine_function, UserDefinedFunction):
            value = combine_function._execute(controller=controller, outcome=outcome, cost=cost, context=context)
        else:
            value = combine_function._execute(variable=[outcome, -cost], context=context)

        return (value, outcome, cost)


class ControlSignalGridSearch(EVCAuxiliaryFunction):
    """Conduct an exhaustive search of allocation polices and return the one with the maximum `EVC
    <EVCControlMechanism_EVC>`.

    This is the default `function <EVCControlMechanism.function>` for an EVCControlMechanism. It identifies the
    `control_allocation` with the maximum `EVC <EVCControlMechanism_EVC>` by a conducting a grid search over every
    possible `control_allocation` given the `allocation_samples` specified for each of its ControlSignals (i.e.,
    the `Cartesian product <https://en.wikipedia.org/wiki/Cartesian_product>`_ of the `allocation
    <ControlSignal.allocation>` values specified by the `allocation_samples` attribute of each ControlSignal).  The
    full set of allocation policies is stored in the EVCControlMechanism's `control_signal_search_space` attribute.
    The EVCControlMechanism's `evaluate` method is used to simulate its `system <EVCControlMechanism.system>`
    under each `control_allocation` in `control_signal_search_space`, calculate the EVC for each of those policies,
    and return the policy with the greatest EVC. By default, only the maximum EVC is saved and returned.  However,
    setting the `save_all_values_and_policies` attribute to `True` saves each policy and its EVC for each simulation
    run (in the EVCControlMechanism's `EVC_policies` and `EVC_values` attributes, respectively). The EVC is
    calculated for each policy by iterating over the following steps:

    * Select an control_allocation:

        draw a successive item from `control_signal_search_space` in each iteration, and assign each of its values as
        the `allocation` value for the corresponding ControlSignal for that simulation of the `system
        <EVCControlMechanism.system>`.

    * Simulate performance:

        execute the `system <EVCControlMechanism.system>` under the selected `control_allocation` using the
        EVCControlMechanism's `evaluate <EVCControlMechanism.evaluate>` method, and the `value
        <PredictionMechanism.value>`\\s of its `prediction_mechanisms <EVCControlMechanism.prediction_mechanisms>` as
        the input to the corresponding `ORIGIN` Mechanisms of the `system <EVCControlMechanism.system>` it controls;
        the values of all :ref:`stateful attributes` of 'Components` in the System are :ref:`re-initialized` to the
        same value prior to each simulation, so that the results for each `control_allocation
        <EVCControlMechanism.control_allocation>` are based on the same initial conditions.  Each simulation includes
        execution of the EVCControlMechanism's `objective_mechanism`, which provides the result to the
        EVCControlMechanism.  If `system <EVCControlMechanism.system>`\\.recordSimulationPref is `True`,
        the results of each simulation are appended to the `simulation_results <System.simulation_results>`
        attribute of `system <EVCControlMechanism.system>`.

    * Calculate the EVC:

        call the EVCControlMechanism's `value_function <EVCControlMechanism_Value_Function>` to calculate the EVC for
        the current iteration, using three values (see `EVCControlMechanism_Functions` for additional details):

        - the EVCControlMechanism's `input <EVC_Mechanism_Input>`, which is the result of its `objective_mechanism
          <EVCControlMechanism.objective_mechanism>`'s `function <ObjectiveMechanism.function>`) and provides an
          evaluation of the outcome of processing in the `system <EVCControlMechanism.system>` under the current
          `control_allocation`;
        |
        - the result of the `cost <EVCControlMechanism_Cost_Function>` function, called by the `value_function
          <EVCControlMechanism_Value_Function>`, that returns the cost for the `control_allocation` based on
          the current `cost <ControlSignal.cost>` associated with each of its ControlSignals;
        |
        - the result of the `combine <EVCControlMechanism_Combine_Function>` function, called by the `value_function
          <EVCControlMechanism_Value_Function>`, that returns the EVC by subtracting the cost from the outcome

    * Save the values:

        if the `save_all_values_and_policies` attribute is `True`, save allocation policy in the EVCControlMechanism's
        `EVC_policies` attribute, and its value is saved in the `EVC_values` attribute; otherwise, retain only
        maximum EVC value.

    The ControlSignalGridSearch function returns the `control_allocation` that yielded the maximum EVC.
    Its operation can be modified by customizing or replacing any or all of the functions referred to above
    (also see `EVCControlMechanism_Functions`).

    """

    componentName = CONTROL_SIGNAL_GRID_SEARCH_FUNCTION

    def __init__(self,
                 default_variable=None,
                 params=None,
                 function=None,
                 owner=None):
        function = function or self.function
        super().__init__(function=function,
                         owner=owner,
                         )

    def _function(
        self,
        controller=None,
        variable=None,
        context=None,
        runtime_params=None,
        params=None,
    ):
        """Grid search combinations of control_signals in specified allocation ranges to find one that maximizes EVC

        * Called by ControlSignalGridSearch.
        * Call System.execute for each `control_allocation` in `control_signal_search_space`.
        * Store an array of values for output_states in `monitored_output_states`
          (i.e., the input_states in `input_states`) for each `control_allocation`.
        * Call `_compute_EVC` for each control_allocation to calculate the EVC, identify the  maximum,
          and assign to `EVC_max`.
        * Set `EVC_max_policy` to the `control_allocation` (outputState.values) corresponding to EVC_max.
        * Set value for each control_signal (outputState.value) to the values in `EVC_max_policy`.
        * Return an control_allocation.

        .. note::
            * runtime_params is used for self.__execute (that calculates the EVC for each call to System.execute);
              it is NOT used for System.execute -- that uses the runtime_params provided for the Mechanisms in each
              Process.configuration

        Return (2D np.array): value of outputState for each monitored state (in self.input_states) for EVC_max

        """

        if self.is_initializing:
            return [defaultControlAllocation]

        # Get value of, or set default for standard args
        if controller is None:
            raise EVCAuxiliaryError("Call to ControlSignalGridSearch() missing controller argument")

        #region RUN SIMULATION

        EVC_max = None
        EVC_values = []
        EVC_policies = []

        # Get allocation_samples for all ControlSignals
        num_control_signals = len(controller.control_signals)
        control_signal_sample_lists = []
        for control_signal in controller.control_signals:
            control_signal_sample_lists.append(control_signal.allocation_samples.generator)
        control_signal_search_space = np.array(np.meshgrid(*control_signal_sample_lists)).T.reshape(-1,num_control_signals)

        # Print progress bar
        if controller.prefs.reportOutputPref:
            progress_bar_rate_str = ""
            search_space_size = len(control_signal_search_space)
            progress_bar_rate = int(10 ** (np.log10(search_space_size)-2))
            if progress_bar_rate > 1:
                progress_bar_rate_str = str(progress_bar_rate) + " "
            print("\n{0} evaluating EVC for {1} (one dot for each {2}of {3} samples): ".
                  format(controller.name, controller.system.name, progress_bar_rate_str, search_space_size))

        # Evaluate all combinations of control_signals (policies)
        sample = 0
        EVC_max_state_values = variable.copy()

        EVC_max_policy = control_signal_search_space[0] * 0.0

        # Parallelize using multiprocessing.Pool
        # NOTE:  currently fails on attempt to pickle lambda functions
        #        preserved here for possible future restoration
        if PY_MULTIPROCESSING:
            EVC_pool = Pool()
            results = EVC_pool.map(compute_EVC, [(controller, arg, runtime_params, context)
                                                 for arg in control_signal_search_space], context=context)

        else:

            # Parallelize using MPI
            if MPI_IMPLEMENTATION:
                Comm = MPI.COMM_WORLD
                rank = Comm.Get_rank()
                size = Comm.Get_size()

                chunk_size = (len(control_signal_search_space) + (size-1)) // size
                print("Rank: {}\nSize: {}\nChunk size: {}".format(rank, size, chunk_size))
                start = chunk_size * rank
                end = chunk_size * (rank+1)
                if start > len(control_signal_search_space):
                    start = len(control_signal_search_space)
                if end > len(control_signal_search_space):
                    end = len(control_signal_search_space)
            else:
                start = 0
                end = len(control_signal_search_space)

            if MPI_IMPLEMENTATION:
                print("START: {0}\nEND: {1}".format(start,end))

            #region EVALUATE EVC

            # Compute EVC for each allocation policy in control_signal_search_space
            # Notes on MPI:
            # * breaks up search into chunks of size chunk_size for each process (rank)
            # * each process computes max for its chunk and returns
            # * result for each chunk contains EVC max and associated allocation policy for that chunk

            result = None
            EVC_max = float('-Infinity')
            EVC_max_policy = np.zeros_like(control_signal_search_space[0])
            EVC_max_state_values = np.zeros_like(controller.get_input_values(context))
            max_value_state_policy_tuple = (EVC_max, EVC_max_state_values, EVC_max_policy)
            # FIX:  INITIALIZE TO FULL LENGTH AND ASSIGN DEFAULT VALUES (MORE EFFICIENT):
            EVC_values = np.array([])
            EVC_policies = np.array([[]])

            # # TEST PRINT EVC:
            # inputs = []
            # for i in controller.predicted_input.values():
            #     inputs.append(repr(i).replace('\n', ''))
            # print("\nEVC SIMULATION for Inputs: {}".format(inputs))

            for allocation_vector in control_signal_search_space[start:end,:]:

            # for iter in range(rank, len(control_signal_search_space), size):
            #     allocation_vector = control_signal_search_space[iter,:]:

                if controller.prefs.reportOutputPref:
                    increment_progress_bar = (progress_bar_rate < 1) or not (sample % progress_bar_rate)
                    if increment_progress_bar:
                        print(kwProgressBarChar, end='', flush=True)
                sample +=1

                # Calculate EVC for specified allocation policy
                result_tuple = compute_EVC(controller, allocation_vector, runtime_params, context)
                EVC, outcome, cost = result_tuple

                EVC_max = max(EVC, EVC_max)
                # max_result([t1, t2], key=lambda x: x1)


                # Add to list of EVC values and allocation policies if save option is set
                if controller.save_all_values_and_policies:
                    # FIX:  ASSIGN BY INDEX (MORE EFFICIENT)
                    EVC_values = np.append(EVC_values, np.atleast_1d(EVC), axis=0)
                    # Save policy associated with EVC for each process, as order of chunks
                    #     might not correspond to order of policies in control_signal_search_space
                    if len(EVC_policies[0])==0:
                        EVC_policies = np.atleast_2d(allocation_vector)
                    else:
                        EVC_policies = np.append(EVC_policies, np.atleast_2d(allocation_vector), axis=0)

                # If EVC is greater than the previous value:
                # - store the current set of monitored state value in EVC_max_state_values
                # - store the current set of control_signals in EVC_max_policy
                # if EVC_max > EVC:
                # FIX: PUT ERROR HERE IF EVC AND/OR EVC_MAX ARE EMPTY (E.G., WHEN EXECUTION_ID IS WRONG)
                if EVC == EVC_max:
                    # Keep track of state values and allocation policy associated with EVC max
                    # EVC_max_state_values = controller.input_value.copy()
                    # EVC_max_policy = allocation_vector.copy()
                    EVC_max_state_values = controller.get_input_values(context)
                    EVC_max_policy = allocation_vector
                    max_value_state_policy_tuple = (EVC_max, EVC_max_state_values, EVC_max_policy)

            # # TEST PRINT EVC:
            # print("EVC_max: {}\tASSOCIATED control_allocation: {}\n".format(EVC_max, EVC_max_policy))

            #endregion

            # Aggregate, reduce and assign global results

            if MPI_IMPLEMENTATION:
                # combine max result tuples from all processes and distribute to all processes
                max_tuples = Comm.allgather(max_value_state_policy_tuple)
                # get tuple with "EVC max of maxes"
                max_of_max_tuples = max(max_tuples, key=lambda max_tuple: max_tuple[0])
                # get EVC_max, state values and allocation policy associated with "max of maxes"
                EVC_max = max_of_max_tuples[0]
                EVC_max_state_values = max_of_max_tuples[1]
                EVC_max_policy = max_of_max_tuples[2]

                if controller.save_all_values_and_policies:
                    EVC_values = np.concatenate(Comm.allgather(EVC_values), axis=0)
                    EVC_policies = np.concatenate(Comm.allgather(EVC_policies), axis=0)

            controller.parameters.EVC_max._set(EVC_max, context)
            controller.parameters.EVC_max_state_values._set(EVC_max_state_values, context)
            controller.parameters.EVC_max_policy._set(EVC_max_policy, context)
            if controller.save_all_values_and_policies:
                controller.parameters.EVC_values._set(EVC_values, context)
                controller.parameters.EVC_policies._set(EVC_policies, context)
            # # TEST PRINT:
            # import re
            # print("\nFINAL:\n\tmax tuple:\n\t\tEVC_max: {}\n\t\tEVC_max_state_values: {}\n\t\tEVC_max_policy: {}".
            #       format(re.sub('[\[,\],\n]','',str(max_value_state_policy_tuple[0])),
            #              re.sub('[\[,\],\n]','',str(max_value_state_policy_tuple[1])),
            #              re.sub('[\[,\],\n]','',str(max_value_state_policy_tuple[2]))),
            #       flush=True)

            # FROM MIKE ANDERSON (ALTERNTATIVE TO allgather:  REDUCE USING A FUNCTION OVER LOCAL VERSION)
            # a = np.random.random()
            # mymax=Comm.allreduce(a, MPI.MAX)
            # print(mymax)

        if controller.prefs.reportOutputPref:
            print("\nEVC simulation completed")
    #endregion

        # -----------------------------------------------------------------

        #region ASSIGN CONTROL SIGNAL VALUES

        # Assign allocations to control_signals for optimal allocation policy:
        EVC_maxStateValue = iter(EVC_max_state_values)

        # Assign max values for optimal allocation policy to controller.input_states (for reference only)
        for i in range(len(controller.input_states)):
            controller.input_states[controller.input_states.names[i]].parameters.value._set(np.atleast_1d(next(EVC_maxStateValue)), context)


        # Report EVC max info
        if controller.prefs.reportOutputPref:
            print ("\nMaximum EVC for {0}: {1}".format(controller.system.name, float(EVC_max)))
            print ("ControlProjection allocation(s) for maximum EVC:")
            for i in range(len(controller.control_signals)):
                print("\t{0}: {1}".format(controller.control_signals[i].name,
                                        EVC_max_policy[i]))
            print()

        #endregion

        # # TEST PRINT:
        # print ("\nEND OF TRIAL 1 EVC outputState: {0}\n".format(controller.outputState.value))

        #region ASSIGN AND RETURN control_allocation
        # Convert EVC_max_policy into 2d array with one control_signal allocation per item,
        #     assign to controller.control_allocation, and return (where it will be assigned to controller.value).
        #     (note:  the conversion is to be consistent with use of controller.value for assignments to control_signals.value)
        allocation_policy = np.array(EVC_max_policy).reshape(len(EVC_max_policy), -1)
        controller.parameters.value._set(allocation_policy, context)
        return allocation_policy
        #endregion


def compute_EVC(ctlr, allocation_vector, runtime_params, context):
    """Compute EVC for a specified `control_allocation <EVCControlMechanism.control_allocation>`.

    IMPLEMENTATION NOTE:  implemented as a function so it can be used with multiprocessing Pool
    IMPLEMENTATION NOTE:  this could be further parallelized if input is for multiple trials

    Simulates and calculates one trial for each set of inputs in ctrl.predicted_input.
    Returns the average EVC over all trials

    Args:
        ctlr (EVCControlMechanism)
        allocation_vector (1D np.array): allocation policy for which to compute EVC
        runtime_params (dict): runtime params passed to ctlr.update
        context (value): context passed to ctlr.update

    Returns (float, float, float):
        (EVC_current, outcome, combined_costs)

    """
    # # TEST PRINT:
    # print("Allocation vector: {}\nPredicted input: {}".
    #       format(allocation_vector, [mech.outputState.value for mech in predicted_input]),
    #       flush=True)


    # Run one simulation and get EVC for each trial's worth of inputs in predicted_input
    predicted_input = ctlr.parameters.predicted_input._get(context)

    origin_mechs = list(predicted_input.keys())
    # number of trials' worth of inputs in predicted_input should be the same for all ORIGIN Mechanisms, so use first:
    num_trials = len(predicted_input[origin_mechs[0]])
    EVC_list = []


    # FIX: 6/16/18: ADD PREDICTION MECHANISM HERE IF IT'S FUNCTION IS STATEFUL
    # Get any values that need to be reinitialized for each run
    reinitialization_values = {}
    for mechanism in ctlr.system.stateful_mechanisms + ctlr.prediction_mechanisms.mechanisms:
        # "save" the current state of each stateful mechanism by storing the values of each of its stateful
        # attributes in the reinitialization_values dictionary; this gets passed into run and used to call
        # the reinitialize method on each stateful mechanism.
        reinitialization_value = []

        if isinstance(mechanism.function, StatefulFunction):
            for attr in mechanism.function.stateful_attributes:
                reinitialization_value.append(mechanism.function.get_current_function_param(attr, context))
        elif hasattr(mechanism, "integrator_function"):
            if isinstance(mechanism.integrator_function, IntegratorFunction):
                for attr in mechanism.integrator_function.stateful_attributes:
                    reinitialization_value.append(mechanism.integrator_function.get_current_function_param(attr, context))

        reinitialization_values[mechanism] = reinitialization_value

    # Run simulation trial by trial in order to get EVC for each trial
    # IMPLEMENTATION NOTE:  Consider calling execute rather than run (for efficiency)
    for i in range(num_trials):
        sim_context = copy.copy(context)
        sim_context.execution_id = ctlr.get_next_sim_id(context)
        # sim_context.add_flag(ContextFlags.SIMULATION)
        try:
            ctlr.parameters.simulation_ids._get(context).append(sim_context.execution_id)
        except AttributeError:
            ctlr.parameters.simulation_ids._set([sim_context.execution_id], context)

        ctlr.system._initialize_from_context(sim_context, context)

        inputs = {key:value[i] for key, value in predicted_input.items()}

        outcome = ctlr.evaluate(
            inputs=inputs,
            allocation_vector=allocation_vector,
            context=sim_context,
            runtime_params=runtime_params,
            reinitialize_values=reinitialization_values,

        )
        EVC_list.append(
            ctlr.value_function(
                controller=ctlr,
                outcome=outcome,
                costs=ctlr.parameters.control_signal_costs._get(sim_context),
                context=sim_context,

            )
        )
        # sim_context.remove_flag(ContextFlags.SIMULATION)
        # assert True
        # # TEST PRINT EVC:
        # print ("Trial: {}\tInput: {}\tAllocation: {}\tOutcome: {}\tCost: {}\tEVC: {}".
        #        format(i, list(inputs.values())[0], allocation_vector,
        #               EVC_list[i][1], EVC_list[i][2], EVC_list[i][0]))

    # Re-assign values of reinitialization attributes to their value at entry
    for mechanism in reinitialization_values:
        mechanism.reinitialize(*reinitialization_values[mechanism], context=context)

    EVC_avg = list(map(lambda x: (sum(x))/num_trials, zip(*EVC_list)))

    # TEST PRINT EVC:
    # print("EVC_avg: {}".format(EVC_avg[0]))

    if PY_MULTIPROCESSING:
        return

    else:
        return (EVC_avg)


INPUT = 'INPUT'
AVERAGE_INPUTS = 'AVERAGE_INPUTS'
INPUT_SEQUENCE = 'INPUT_SEQUENCE'
TIME_AVERAGE_INPUT = 'TIME_AVERAGE_INPUT'
input_types = {INPUT, TIME_AVERAGE_INPUT, AVERAGE_INPUTS, INPUT_SEQUENCE}

WINDOW_SIZE = 'window_size'
FILTER_FUNCTION = 'filter_function'


class PredictionMechanism(IntegratorMechanism):
    """PredictionMechanism(      \
    default_variable=None,       \
    size=None,                   \
    function=TIME_AVERAGE_INPUT, \
    initial_value=None,          \
    window_size=1,               \
    filter_function=None,        \
    params=None,                 \
    name=None,                   \
    prefs=None)

    Tracks the inputs to an `ORIGIN` Mechanism of the `system <EVCControlMechanism.system>` controlled by an
    `EVCControlMechanism`, and used to generate the input for that `ORIGIN` Mechanism in a `simulated run
    <EVCControlMechanism_Execution>` of that System.

    .. _PredictionMechanism_Creation:

    **Creating a PredictionMechanism**

    PredictionMechanisms are created automatically when an `EVCControlMechanism` is created, one for each `ORIGIN`
    Mechanism in the `system <EVCControlMechanism.system>` for which the EVCControlMechanism is a `controller
    <System.controller>`. PredictionMechanisms should not be created on their own, as their execution requires
    tight integration with an EVCControlMechanism and the System to which it belongs, and they will not function
    properly if this is not insured.

    **Structure**

    The `System` to which a PredictionMechanism belongs is referenced in its `system <PredictionMechanism.system>`
    attribute, and the System's `ORIGIN` Mechanism which the PredictionMechanism is associated is referenced in its
    `origin_mechanism <PredictionMechanism.origin_mechanism>` attribute.  A PredictionMechanism has the same number
    of `InputStates <InputState>` as its `origin_mechanism <PredictionMechanism.origin_mechanism>`, each of which
    receives a `Projection <Projection>` from the same source as the corresponding InputState of its `origin_mechanism
    <PredictionMechanism.origin_mechanism>` (see `EVCControlMechanism_Prediction_Mechanisms`); and it has one
    `OutputState` for each of its InputStates.

    .. _PredictionMechanism_Function:

    **Function**

    The `function <PredictionMechanism.function>` of a PredictionMechanism records the input received by its
    `origin_mechanism <PredictionMechanism.origin_mechanism>`, and possibly transforms this in some way.  Any
    function can be assigned, so long as it can take as its input a value with the same format as the `variable
    <Mechanism_Base.variable>` of its `origin_mechanism <PredictionMechanism.origin_mechanism>`, and returns a
    similarly formatted value or a list of them.  The result of a PredictionMechanism's `function
    <PredictionMechanism.function>` is provided as the input to its `origin_mechanism
    <PredictionMechanism.origin_mechanism>` in each trial of a simulated run of its `system
    <PredictionMechanism.system>` (see `EVCControlMechanism Execution <EVCControlMechanism_Execution>`).  If a
    PredictionMechanism's `function <PredictionMechanism.function>` returns more than one item,
    then the EVCControlMechanism runs as many simulations as there are items, using each as the input for one
    simulation, and computes the mean EVC over all of them.  Therefore, **the** `function
    <PredictionMechanism.function>` **of every PredictionMechanism associated with an** `EVCControlMechanism`
    **must return the same number of items.**

    In place of a function, the following keywords can be used to specify one of three standard configurations:

    * *INPUT:* uses an `Linear` Function to simply pass the input it receives as its output, possibly modified by the
      values specified in the PredictionMechanism's **rate** and **noise** arguments.  If those are specified, they
      are assigned as the `Linear` function's `slope <Linear.slope>` and `intercept <Linear.intercept>` parameters,
      respectively.

    * *TIME_AVERAGE_INPUT:* uses an `AdaptiveIntegrator` Function to compute an exponentially weighted
      time-average of the input to the PredictionMechanism; the PredictionMechanism's **rate** and **noise**
      arguments can be used to specify the corresponding `rate <AdaptiveIntegrator.rate>` and `noise
      <AdaptiveIntegrator.noise>` parameters of the function.  The function returns the time-averaged input
      as a single item.

    * *AVERAGE_INPUTS:* uses a `Buffer` Function to compute the average of the number of preceding inputs specified in
      the PredictionMechanism's **window_size** argument.  If the **rate** and/or **noise** arguments are specified,
      they are applied to each item in the list before it is averaged, as follows: :math:`item * rate + noise`.

    .. _PredictionMechanism_Input_Sequence:

    * *INPUT_SEQUENCE:* uses a `Buffer` Function to maintain a running record of preceding inputs, which are returned
      in a list. When the EVCControlMechanism `runs a simulation <EVCControlMechanism_Execution>`, one trial is run
      using each item in the list as the input to the PredictionMechanism's `origin_mechanism
      <PredictionMechanism.origin_mechanism>`. The **window_size** argument can be used to specify how many preceding
      inputs should be maintained in the record; if the number of preceding inputs exceeds the value of
      **window_size**, the oldest is deleted and the most recent one is added to the list.  If **window_size** is not
      specified (`None`), a record of all preceding inputs is maintained and returned. The **filter_function**
      argument can be used to specify a function that filters the list (e.g., modifies and/or deletes items);  the
      modified list is then returned by the PredictionMechanism's `function <PredictionMechanism.function>`.  If the
      **rate** and/or **noise** arguments are specified, they are applied to each item in the list, as follows:
      :math:`item * rate + noise`;  note that since the list is maintained across trials in the actual run of the
      `system <PredictionMechanism.system>`, the effects of these parameters are cumulative over trials.

    **Execution**

    A PredictionMechanism is executed each time its `system <PredictionMechanism>` is run; however, it is **not**
    executed when that System is simulated (that is, it is run using the EVCControlMechanism's `evaluate
    <EVCControlMechanism.evaluate>` method).  Thus, its inputs are updated only once per *actual* run of the
    System, and each simulated run (i.e., the simulation for each `control_allocation
    <EVCControlMechanism.control_allocation>`; see `EVCControlMechanism Execution <EVCControlMechanism_Execution>`)
    uses the *exact same* set of inputs -- the value of the PredictionMechanisms resulting from the last actual run
    of its `system <PredictionMechanism.system>`) -- so that the results of the simulated run for each
    `control_allocation <EVCControlMechanism.control_allocation>` can properly be compared. If the PredictionMechanisms
    for an EVCControlMechanism generate a list of inputs (e.g., using `INPUT_SEQUENCE
    <PredictionMechanism_Input_Sequence>`), then each simulated run involves several trials, each of which uses one
    item from the list of each PredictionMechanism as the input to its `origin_mechanism
    <PredictionMechanism.origin_mechanism>`;  items in the list are used as inputs in the exact same sequence for the
    simulated run of every `control_allocation <EVCControlMechanism.control_allocation>` so that, once again,
    proper comparisons can be made between policies.

    Arguments
    ---------

    default_variable : number, list or np.ndarray : default None
        specifies the input to the Mechanism to use if none is provided in a call to its
        `execute <Mechanism_Base.execute>` or `run <Mechanism_Base.run>` method; format must match the
        `variable <Mechanism.variable>` for its `origin_mechanism <PredictionMechanism.origin_mechanism>`.

    size : int, list or np.ndarray of ints
        see `size <Mechanism.size>`.

    function : function, *INPUT*, *TIME_AVERAGE_INPUT*, *AVERAGE_INPUTS*, or *INPUT_SEQUENCE* : default
    *TIME_AVERAGE_INPUT*
        specifies the function used to generate the input provided to its `origin_mechanism
        <PredictionMechanism.origin_mechanism>`; the function must take as its input a single value with the same
        format as the `variable <Mechanism.variable>` of its `origin_mechanism <PredictionMechanism.origin_mechanism>`,
        and must return a similarly formatted value or a list of them (see `above <PredictionMechanism_Function>`
        for additional details).

    initial_value :  value, list or np.ndarray : default None
        specifies value used to initialize the PredictionMechanism's `value <PredictionMechanism.value>` attribute;
        if `None` is specified, 0 is used if the `value <Function_Base.value>` of the PredictionMechanism's `function
        <PredictionMechanism.function>` is numeric, and an empty list is used if *INPUT_SEQUENCE* is specified.

    window_size : int : default None
        specifies number of input values to maintain when *INPUT_SEQUENCE* option is used for
        `function <PredictionMechanism.function>`

    filter_function: function : default None
        specifies a function that takes a list of values, each of which has the same format as the `variable
        <Mechanism_Base.variable>` of the PredictionMechanism's `origin_mechanism
        <PredictionMechanism.origin_mechanism>`, and returns a list of similarly formatted values, though not
        necessarily of the same length (see `above <PredictionMechanism_Function>` for additional details).

    params : Dict[param keyword: param value] : default None
        a `parameter dictionary <ParameterState_Specification>` that can be used to specify the parameters for
        the Mechanism, its `function <Mechanism_Base.function>`, and/or a custom function and its parameters.  Values
        specified for parameters in the dictionary override any assigned to those parameters in arguments of the
        constructor.

    name : str : default see `name <TransferMechanism.name>`
        specifies the name of the TransferMechanism.

    prefs : PreferenceSet or specification dict : default Mechanism.classPreferences
        specifies the `PreferenceSet` for the TransferMechanism; see `prefs <TransferMechanism.prefs>` for details.

    Attributes
    ----------

    system : System
        the `System` to which the PredictionMechanism belongs.

    origin_mechanism : Mechanism
        the `ORIGIN` Mechanism with which the PredictionMechanism is associated (in its `system
        <PredictionMechanism.system>`), and the input of which it tracks.

    variable :  2d np.array
        the input received from the PredictionMechanism's `system <PredictionMechanism.system>`, that is identical
        to the input received by its `origin_mechanism <PredictionMechanism.origin_mechanism>`.

    function : Function
        used to track the inputs to the PredictionMechanism's `origin_mechanism <PredictionMechanism.origin_mechanism>`;
        the default is an `AdaptiveIntegrator` (see `above <PredictionMechanism_Function>` for additional details).

    value : 3d np.array
        result returned by the PredictionMechanism's `function <PredictionMechanism.function>`, and provided as
        input to its `origin_mechanism <PredictionMechanism.origin_mechanism>`.  The format conforms to that of a
        System's `run <System.run>` method: items in the outermost dimension (axis 0) correspond to the inputs for
        each trial of a simulation, each of which is a 2d np.array containing the input for each `InputState` of the
        `Mechanism`.

    """

    componentType = PREDICTION_MECHANISM

    class Parameters(IntegratorMechanism.Parameters):
        """
            Attributes
            ----------

                filter_function
                    see `filter_function <PredictionMechanism.filter_function>`

                    :default value: None
                    :type:

                rate
                    see `rate <PredictionMechanism.rate>`

                    :default value: 1.0
                    :type: float

                window_size
                    see `window_size <PredictionMechanism.window_size>`

                    :default value: 1
                    :type: int

        """
        window_size = Parameter(1, stateful=False, loggable=False)
        filter_function = Parameter(None, stateful=False, loggable=False)

        rate = Parameter(1.0, modulable=True)

    @tc.typecheck
    @handle_external_context(source=None)
    def __init__(self,
                 default_variable=None,
                 size=None,
                 input_states:tc.optional(tc.any(list, dict))=None,
                 function:tc.optional(tc.enum(
                         INPUT, TIME_AVERAGE_INPUT, AVERAGE_INPUTS, INPUT_SEQUENCE))=TIME_AVERAGE_INPUT,
                 initial_value=None,
                 # rate:tc.optional(tc.any(int, float))=1.0,
                 # noise:tc.optional(tc.any(int, float, callable))=0.0,
                 rate:tc.any(int, float)=1.0,
                 noise:tc.any(int, float, callable)=0.0,
                 window_size=1,
                 filter_function:tc.optional(callable)=None,
                 params=None,
                 name=None,
                 prefs:is_pref_set=None,
                 context=None):

        if not context.source in {ContextFlags.COMPONENT, ContextFlags.COMPOSITION, ContextFlags.COMMAND_LINE}:
            warnings.warn("PredictionMechanism should not be constructed on its own.  If you insist,"
                          "set context=Context(source=ContextFlags.COMMAND_LINE), but proceed at your peril!")
            return

        if params and FUNCTION in params:
            function = params[FUNCTION]

        input_type = None
        if function in input_types:
            input_type = function

        params = self._assign_args_to_param_dicts(window_size=window_size,
                                                  input_type=input_type,
                                                  filter_function=filter_function,
                                                  params=params)

        if function in input_types:

            if function is INPUT:
                function = Linear(slope=rate, intercept=noise)

            elif function is TIME_AVERAGE_INPUT:
                # Use default for IntegratorMechanism: AdaptiveIntegrator
                function = self.class_defaults.function

            elif function in {AVERAGE_INPUTS, INPUT_SEQUENCE}:

                # Maintain the preceding sequence of inputs (of length window_size), and use those for each simulation
                function = Buffer(default_variable=[[0]],
                                  initializer=initial_value,
                                  rate=rate,
                                  noise=noise,
                                  history=window_size)

        params.update({FUNCTION_PARAMS:{RATE:rate,
                                        NOISE:noise}})

        super().__init__(
                default_variable=default_variable,
                size=size,
                input_states=input_states,
                function=function,
                params=params,
                name=name,
                prefs=prefs)

    def _execute(self, variable=None, context=None, runtime_params=None):
        """Update predicted value on "real" but not simulation runs"""

        if ContextFlags.SIMULATION in context.execution_phase:
            # Just return current value for simulation runs
            value = self.parameters.value._get(context)

        else:
            # Update deque with new input for any other type of run
            value = super()._execute(variable=variable, context=context, runtime_params=runtime_params)

            # If inputs are being recorded (#recorded = window_size):
            if len(value) > 1:
                if self.input_type is AVERAGE_INPUTS:
                    # Compute average input over window_size
                    value = np.sum(value)/value.shape[0]

                elif self.input_type is INPUT_SEQUENCE:
                    if self.filter_function:
                        # Use filter_function to return input values
                        value = self.filter_function(value)
                    else:
                        # Return all input values in window_size
                        pass
        return value

    @property
    def system(self):
        try:
            from psyneulink.core.components.system import System
            return next((p.sender.owner for p in self.afferents if isinstance(p.sender.owner, System)), None)
        except:
            return self._system

    @system.setter
    def system(self, value):
        self._system = value

    @property
    def origin_mechanism(self):
        try:
            return self.system.origin_mechanisms[self.system.controller.prediction_mechanisms.index(self)]
        except:
            return None





# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
#
#
# *************************************************  Utilities *********************************************************

"""Utilities that must be accessible to all PsyNeuLink modules, but are not PsyNeuLink-specific

   That is:
       do not require any information about PsyNeuLink objects
       are not constrained to be used by PsyNeuLink objects

************************************************* UTILITIES ************************************************************


CONTENTS
--------

*TYPE CHECKING VALUE COMPARISON*
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

.. note::
   PsyNeuLink-specific typechecking functions are in the `Component <Component>` module

* `parameter_spec`
* `optional_parameter_spec`
* `all_within_range`
* `is_matrix
* `is_matrix_spec`
* `is_numeric`
* `is_numeric_or_none`
* `is_iterable`
* `iscompatible`
* `is_value_spec`
* `is_unit_interval`
* `is_same_function_spec`
* `is_component`

*ENUM*
~~~~~~

* `Autonumber`
* `Modulation`
* `get_modulationOperation_name`

*KVO*
~~~~~

.. note::
   This is for potential future use;  not currently used by PsyNeuLink objects

* observe_value_at_keypath

*MATHEMATICAL*
~~~~~~~~~~~~~~

* norm
* sinusoid
* scalar_distance
* powerset
* tensor_power


*OTHER*
~~~~~~~

* `get_args`
* `recursive_update`
* `merge_param_dicts`
* `multi_getattr`
* `np_array_less_that_2d`
* `convert_to_np_array`
* `type_match`
* `get_value_from_array`
* `is_matrix`
* `underscore_to_camelCase`
* `append_type_to_name`
* `ReadOnlyOrderedDict`
* `ContentAddressableList`
* `make_readonly_property`
* `get_class_attributes`
* `insert_list`
* `flatten_list`
* `convert_to_list`
* `get_global_seed`
* `set_global_seed`

"""

import copy
import inspect
import logging
import numbers
import time
import warnings
import weakref

from enum import Enum, EnumMeta, IntEnum

import collections
import numpy as np

from psyneulink.core.globals.keywords import DISTANCE_METRICS, EXPONENTIAL, GAUSSIAN, LINEAR, MATRIX_KEYWORD_VALUES, NAME, SINUSOID, VALUE

__all__ = [
    'append_type_to_name', 'AutoNumber', 'ContentAddressableList', 'convert_to_list', 'convert_to_np_array',
    'convert_all_elements_to_np_array', 'NodeRole', 'get_class_attributes', 'flatten_list',
    'get_modulationOperation_name', 'get_value_from_array', 'is_component',
    'is_distance_metric', 'is_matrix',
    'insert_list', 'is_matrix_spec', 'all_within_range', 'is_iterable',
    'is_modulation_operation', 'is_numeric', 'is_numeric_or_none', 'is_same_function_spec', 'is_unit_interval',
    'is_value_spec', 'iscompatible', 'kwCompatibilityLength', 'kwCompatibilityNumeric', 'kwCompatibilityType',
    'make_readonly_property', 'merge_param_dicts',
    'Modulation', 'MODULATION_ADD', 'MODULATION_MULTIPLY','MODULATION_OVERRIDE',
    'multi_getattr', 'np_array_less_than_2d', 'object_has_single_value', 'optional_parameter_spec', 'normpdf',
    'parameter_spec', 'powerset', 'random_matrix', 'ReadOnlyOrderedDict', 'safe_len', 'scalar_distance', 'sinusoid',
    'tensor_power', 'TEST_CONDTION', 'type_match',
    'underscore_to_camelCase', 'UtilitiesError', 'unproxy_weakproxy'
]

logger = logging.getLogger(__name__)


class UtilitiesError(Exception):
    def __init__(self, error_value):
        self.error_value = error_value

    def __str__(self):
        return repr(self.error_value)


MODULATION_OVERRIDE = 'Modulation.OVERRIDE'
MODULATION_MULTIPLY = 'Modulation.MULTIPLY'
MODULATION_ADD = 'Modulation.ADD'

class Modulation(Enum):
    MULTIPLY = lambda runtime, default : runtime * default
    ADD = lambda runtime, default : runtime + default
    OVERRIDE = lambda runtime, default : runtime
    DISABLE = 0

def is_modulation_operation(val):
    # try:
    #     val(0,0)
    #     return True
    # except:
    #     return False
    return get_modulationOperation_name(val)

def get_modulationOperation_name(operation):
        x = operation(1,2)
        if x is 1:
            return MODULATION_OVERRIDE
        elif x is 2:
            return MODULATION_MULTIPLY
        elif x is 3:
            return MODULATION_ADD
        else:
            return False



class AutoNumber(IntEnum):
    """Autonumbers IntEnum type

    First item in list of declarations is numbered 0;
    Others incremented by 1

    Sample:

        >>> class NumberedList(AutoNumber):
        ...    FIRST_ITEM = ()
        ...    SECOND_ITEM = ()

        >>> NumberedList.FIRST_ITEM.value
        0
        >>> NumberedList.SECOND_ITEM.value
        1

    Adapted from AutoNumber example for Enum at https://docs.python.org/3/library/enum.html#enum.IntEnum:
    Notes:
    * Start of numbering changed to 0 (from 1 in example)
    * obj based on int rather than object
    """
    def __new__(component_type):
        # Original example:
        # value = len(component_type.__members__) + 1
        # obj = object.__new__(component_type)
        value = len(component_type.__members__)
        obj = int.__new__(component_type)
        obj._value_ = value
        return obj

# ******************************** GLOBAL STRUCTURES, CONSTANTS AND METHODS  *******************************************

TEST_CONDTION = False


def optional_parameter_spec(param):
    """Test whether param is a legal PsyNeuLink parameter specification or `None`

    Calls parameter_spec if param is not `None`
    Used with typecheck

    Returns
    -------
    `True` if it is a legal parameter or `None`.
    `False` if it is neither.


    """
    if not param:
        return True
    return parameter_spec(param)


def parameter_spec(param, numeric_only=None):
    """Test whether param is a legal PsyNeuLink parameter specification

    Used with typecheck

    Returns
    -------
    `True` if it is a legal parameter.
    `False` if it is not.
    """
    # if isinstance(param, property):
    #     param = ??
    # if is_numeric(param):
    from psyneulink.core.components.functions.function import function_type
    from psyneulink.core.components.shellclasses import Projection
    from psyneulink.core.components.component import parameter_keywords
    from psyneulink.core.globals.keywords import MODULATORY_SPEC_KEYWORDS
    from psyneulink.core.components.component import Component

    if inspect.isclass(param):
        param = param.__name__
    elif isinstance(param, Component):
        param = param.__class__.__name__
    if (isinstance(param, (numbers.Number,
                           np.ndarray,
                           list,
                           tuple,
                           dict,
                           function_type,
                           Projection))
        or param in MODULATORY_SPEC_KEYWORDS
        or param in parameter_keywords):
        if numeric_only:
            if not is_numeric(param):
                return False
        return True
    return False


def all_within_range(x, min, max):
    x = np.array(x)
    try:
        if min is not None and (x<min).all():
            return False
        if max is not None and (x>max).all():
            return False
        return True
    except (ValueError, TypeError):
        for i in x:
            if not all_within_range(i, min, max):
                return False
        return True

def is_numeric_or_none(x):
    if x is None:
        return True
    return is_numeric(x)


def is_numeric(x):
    return iscompatible(x, **{kwCompatibilityNumeric:True, kwCompatibilityLength:0})


def is_matrix_spec(m):
    return isinstance(m, str) and m in MATRIX_KEYWORD_VALUES


def is_matrix(m):
    from psyneulink.core.components.component import Component

    if is_matrix_spec(m):
        return True
    if isinstance(m, (list, np.ndarray, np.matrix)):
        return True
    if m is None or isinstance(m, (Component, dict, set)) or (inspect.isclass(m) and issubclass(m, Component)):
        return False
    try:
        m2 = np.matrix(m)
        return is_matrix(m2)
    except:
        pass
    if callable(m):
        try:
            return is_matrix(m())
        except:
            return False
    return False


def is_distance_metric(s):
    if s in DISTANCE_METRICS:
        return True
    else:
        return False


def is_iterable(x):
    """
    Returns
    -------
        True - if **x** can be iterated on
        False - otherwise
    """
    if isinstance(x, np.ndarray) and x.ndim == 0:
        return False
    else:
        return isinstance(x, collections.abc.Iterable)


kwCompatibilityType = "type"
kwCompatibilityLength = "length"
kwCompatibilityNumeric = "numeric"

# IMPLEMENT SUPPORT OF *LIST* OF TYPES IN kwCompatibilityType (see Function.__init__ FOR EXAMPLE)
# IMPLEMENT: IF REFERENCE IS np.ndarray, try converting candidate to array and comparing
def iscompatible(candidate, reference=None, **kargs):
    """Check if candidate matches reference or, if that is omitted, it matches type, length and/or numeric specification

    If kargs is omitted, candidate and reference must match in type and length
    If reference and kargs are omitted, candidate must be a list of numbers (of any length)
    If reference is omitted, candidate must match specs in kargs

    kargs is an optional dictionary with the following entries:
        kwCompatibilityType ("type"):<type> (default: list):  (spec local_variable: match_type)
            - if reference is provided, candidate's type must match or be subclass of reference,
                irrespective of whether kwCompatibilityType is specified or absent;  however, if:
                + kwCompatibilityType is absent, enums are treated as numbers (of which they are a subclass)
                + kwCompatibilityType = enum, then candidate must be an enum if reference is one
            - if reference is absent:
                if kwCompatibilityType is also absent:
                    if kwCompatibilityNumeric is `True`, all elements of candidate must be numbers
                    if kwCompatibilityNumeric is :keyword:`False`, candidate can contain any type
                if kwCompatibilityType is specified, candidate's type must match or be subclass of specified type
            - for iterables, if kwNumeric is :keyword:`False`, candidate can have multiple types but
                if a reference is provided, then the corresponding items must have the same type
        kwCompatibilityLength ("length"):<int>  (default: 0):    (spec local_variable: match_length)
            - if kwCompatibilityLength is absent:
                if reference is absent, candidate can be of any length
                if reference is provided, candidate must match reference length
            - if kwCompatibilityLength = 0:
                candidate can be any length, irrespective of the reference or its length
            - if kwCompatibilityLength > 0:
                if reference is provided, candidate must be same length as reference
                if reference is omitted, length of candidate must equal value of kwLength
            Note: kwCompatibility < 0 is illegal;  it will generate a warning and be set to 0
        kwCompatibilityNumeric ("number": <bool> (default: `True`)  (spec local_variable: number_only)
            If kwCompatibilityNumeric is `True`, candidate must be either numeric or a list or tuple of
                numeric types
            If kwCompatibilityNumberic is :keyword:`False`, candidate can be strings, lists or tuples of strings,
                or dicts
                Note: if the candidate is a dict, the number of entries (lengths) are compared, but not their contents

    :param candidate: (value)
    :param reference:  (value)
    :param args: (dict)
    :return:
    """

    # If the two are equal, can settle it right here
    # IMPLEMENTATION NOTE: remove the duck typing when numpy supports a direct comparison of iterables
    try:
        with warnings.catch_warnings():
            warnings.simplefilter(action='ignore', category=FutureWarning)
            if reference is not None and (candidate == reference):
                return True
    except ValueError:
        # raise UtilitiesError("Could not compare {0} and {1}".format(candidate, reference))
        # IMPLEMENTATION NOTE: np.array generates the following error:
        # ValueError: The truth value of an array with more than one element is ambiguous. Use a.any() or a.all()
        pass

    # If args not provided, assign to default values
    # if not specified in args, use these:
    #     args[kwCompatibilityType] = list
    #     args[kwCompatibilityLength] = 1
    #     args[kwCompatibilityNumeric] = True
    try:
        kargs[kwCompatibilityType]
    except KeyError:
        kargs[kwCompatibilityType] = list
    try:
        kargs[kwCompatibilityLength]
    except KeyError:
        kargs[kwCompatibilityLength] = 1
    try:
        # number_only = kargs[kwCompatibilityNumeric]
        kargs[kwCompatibilityNumeric]
    except KeyError:
        kargs[kwCompatibilityNumeric] = True
        # number_only = True


    # If reference is not provided, assign local_variables to arg values (provided or default)
    if reference is None:
        match_type = kargs[kwCompatibilityType]
        match_length = kargs[kwCompatibilityLength]
        number_only = kargs[kwCompatibilityNumeric]
    # If reference is provided, assign specification local_variables to reference-based values
    else:
        match_type = type(reference)
        # If length specification is non-zero (i.e., use length) and reference is an object for which len is defined:
        if (kargs[kwCompatibilityLength] and
                (isinstance(reference, (list, tuple, dict)) or
                         isinstance(reference, np.ndarray) and reference.ndim)
            ):
            match_length = len(reference)
        else:
            match_length = 0
        # If reference is not a number, then don't require the candidate to be one
        if not isinstance(reference, numbers.Number):
            number_only = False
        else:
            number_only = kargs[kwCompatibilityNumeric]

    if match_length < 0:
        # if settings & Settings.VERBOSE:
        print("\niscompatible({0}, {1}): length argument must be non-negative; it has been set to 0\n".
              format(candidate, kargs, match_length))
        match_length = 0

    # # FIX??
    # # Reference is a matrix or a keyword specification for one
    if is_matrix_spec(reference):
        return is_matrix(candidate)

    # MODIFIED 10/29/17 NEW:
    # IMPLEMENTATION NOTE: This allows a number in an ndarray to match a float or int
    # If both the candidate and reference are either a number or an ndarray of dim 0, consider it a match
    if ((isinstance(candidate, numbers.Number) or (isinstance(candidate, np.ndarray) and candidate.ndim == 0)) or
            (isinstance(reference, numbers.Number) or (isinstance(reference, np.ndarray) and reference.ndim == 0))):
        return True
    # MODIFIED 10/29/17 END

    # IMPLEMENTATION NOTE:
    #   modified to allow numeric type mismatches (e.g., int and float;
    #   should be added as option in future (i.e., to disallow it)
    if (isinstance(candidate, match_type) or
            (isinstance(candidate, (list, np.ndarray)) and (issubclass(match_type, (list, np.ndarray)))) or
            (isinstance(candidate, numbers.Number) and issubclass(match_type,numbers.Number)) or
            # IMPLEMENTATION NOTE: Allow UserDict types to match dict (make this an option in the future)
            (isinstance(candidate, UserDict) and match_type is dict) or
            # IMPLEMENTATION NOTE: Allow UserList types to match list (make this an option in the future)
            (isinstance(candidate, UserList) and match_type is list) or
            # IMPLEMENTATION NOTE: This is needed when kwCompatiblityType is not specified
            #                      and so match_type==list as default
            (isinstance(candidate, numbers.Number) and issubclass(match_type,list)) or
                (isinstance(candidate, np.ndarray) and issubclass(match_type,list))
        ):

        # Check compatibility of enum's
        # IMPLEMENTATION NOTE: THE FIRST VERSION BELOW SOUGHT TO CHECK COMPATIBILTY OF ENUM VALUE;  NEEDS WORK
        # if (kargs[kwCompatibilityType] is Enum and
        #         (isinstance(candidate, Enum) or isinstance(match_type, (Enum, IntEnum, EnumMeta)))):
        #     # If either the candidate enum's label is not in the reference's Enum class
        #     #    or its value is different, then return with fail
        #     try:
        #         match_type.__dict__['_member_names_']
        #     except:
        #         pass
        #     if not (candidate.name in match_type.__dict__['_member_names_'] and
        #                 candidate.value is match_type.__dict__['_member_map_'][candidate.name].value):
        #         return False
        # This version simply enforces the constraint that, if either is an enum of some sort, then both must be
        if (kargs[kwCompatibilityType] is Enum and
                (isinstance(candidate, Enum) != isinstance(match_type, (Enum, IntEnum, EnumMeta)))
            ):
            return False

        if isinstance(candidate, numbers.Number):
            return True
        if number_only:
            if isinstance(candidate, np.ndarray) and candidate.ndim ==0 and np.isreal(candidate):
                return True
            if not isinstance(candidate, (list, tuple, np.ndarray, np.matrix)):
                return False
            def recursively_check_elements_for_numeric(value):
                # Matrices can't be checked recursively, so convert to array
                if isinstance(value, np.matrix):
                    value = value.A
                if isinstance(value, (list, np.ndarray)):
                    for item in value:
                        if not recursively_check_elements_for_numeric(item):
                            return False
                        else:
                            return True
                else:
                    if not isinstance(value, numbers.Number):
                        try:
                            # True for autograd ArrayBox (and maybe other types?)
                            # if isinstance(value._value, numbers.Number):
                            from autograd.numpy.numpy_boxes import ArrayBox
                            if isinstance(value, ArrayBox):
                                return True
                        except:
                            return False
                    else:
                        return True
            # Test copy since may need to convert matrix to array (see above)
            if not recursively_check_elements_for_numeric(candidate.copy()):
                return False

        if isinstance(candidate, (list, tuple, dict, np.ndarray)):
            if not match_length:
                return True
            else:
                if len(candidate) == match_length:
                    # No reference, so item by item comparison is not relevant
                    if reference is None:
                        return True
                    # Otherwise, carry out recursive elementwise comparison
                    if isinstance(candidate, np.matrix):
                        candidate = np.asarray(candidate)
                    if isinstance(reference, np.matrix):
                        reference = np.asarray(reference)
                    cr = zip(candidate, reference)
                    if all(iscompatible(c, r, **kargs) for c, r in cr):
                        return True
                    # IMPLEMENTATION NOTE:  ??No longer needed given recursive call above
                    # Deal with ints in one and floats in the other
                    # # elif all((isinstance(c, numbers.Number) and isinstance(r, numbers.Number))
                    # #          for c, r in cr):
                    # #     return True
                else:
                    return False
        else:
            return True
    else:
        return False


# MATHEMATICAL  ********************************************************************************************************

def normpdf(x, mu=0, sigma=1):
    u = float((x-mu) / abs(sigma))
    y = np.exp(-u*u/2) / (np.sqrt(2*np.pi) * abs(sigma))
    return y

def sinusoid(x, amplitude=1, frequency=1, phase=0):
    return amplitude * np.sin(2 * np.pi * frequency * x + phase)

def scalar_distance(measure, value, scale=1, offset=0):
    if measure is GAUSSIAN:
        return normpdf(value, offset, scale)
    if measure is LINEAR:
        return scale*value+offset
    if measure is EXPONENTIAL:
        return np.exp(scale*value+offset)
    if measure is SINUSOID:
        return sinusoid(value, frequency=scale, phase=offset)

from itertools import chain, combinations
def powerset(iterable):
    """powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)"""
    s = list(iterable)
    return chain.from_iterable(combinations(s, r) for r in range(len(s)+1))

import typecheck as tc
@tc.typecheck
def tensor_power(items, levels:tc.optional(range)=None, flat=False):
    """return tensor product for all members of powerset of items

    levels specifies a range of set levels to return;  1=first order terms, 2=2nd order terms, etc.
    if None, all terms will be returned

    if flat=False, returns list of 1d arrays with tensor product for each member of the powerset
    if flat=True, returns 1d array of values
    """

    ps = list(powerset(items))
    max_levels = max([len(s) for s in ps])
    levels = levels or range(1,max_levels)
    max_spec = max(list(levels))
    min_spec = min(list(levels))
    if  max_spec > max_levels:
        raise UtilitiesError("range ({},{}) specified for {} arg of tensor_power() "
                             "exceeds max for items specified ({})".
                             format(min_spec, max_spec+1, repr('levels'), max_levels+1))

    pp = []
    for s in ps:
        order = len(s)
        if not order in list(levels):
            continue
        if order==1:
            pp.append(np.array(s[0]))
        else:
            i = 0
            tp = np.tensordot(s[i],s[i+1],axes=0)
            i+=2
            while i < order:
                tp = np.tensordot(tp, s[i], axes=0)
                i+=1
            if flat is True:
                pp.extend(tp.reshape(-1))
            else:
                pp.append(tp.reshape(-1))
    return pp



# OTHER ****************************************************************************************************************

def get_args(frame):
    """Gets dictionary of arguments and their values for a function
    Frame should be assigned as follows in the function itself:  frame = inspect.currentframe()
    """
    args, _, _, values = inspect.getargvalues(frame)
    return dict((key, value) for key, value in values.items() if key in args)


from collections.abc import Mapping
def recursive_update(d, u, non_destructive=False):
    """Recursively update entries of dictionary d with dictionary u
    From: https://stackoverflow.com/questions/3232943/update-value-of-a-nested-dictionary-of-varying-depth
    """
    for k, v in u.items():
        if isinstance(v, Mapping):
            r = recursive_update(d.get(k, {}), v)
            d[k] = r
        else:
            if non_destructive and k in d and d[k] is not None:
                continue
            d[k] = u[k]
    return d


def merge_param_dicts(source, specific, general):
    """Search source dict for specific and general dicts, merge specific with general, and return merged

    Description:
        - used to merge only a subset of dicts in param set (that may have several dicts
        - allows dicts to be referenced by name (e.g., paramName) rather than by object
        - searches source dict for specific and general dicts
        - if both are found, merges them, with entries from specific overwriting any duplicates in general
        - if only one is found, returns just that dict
        - if neither are found, returns empty dict

    Arguments:
        - source (dict): container dict (entries are dicts); search entries for specific and general dicts
        - specific (dict or str): if str, use as key to look for specific dict in source, and check that it is a dict
        - general (dict or str): if str, use as key to look for general dict in source, and check that it is a dict


    :param source: (dict)
    :param specific: (dict or str)
    :param general: (dict or str)
    :return merged: (dict)
    """

    # Validate source as dict
    if not source:
        return
    if not isinstance(source, dict):
        raise UtilitiesError("merge_param_dicts: source {0} must be a dict".format(source))

    # Get specific and make sure it is a dict
    if isinstance(specific, str):
        try:
            specific = source[specific]
        except (KeyError, TypeError):
            specific = {}
    if not isinstance(specific, dict):
        raise UtilitiesError("merge_param_dicts: specific {0} must be dict or the name of one in {1}".
                        format(specific, source))

    # Get general and make sure it is a dict
    if isinstance(general, str):
        try:
            general = source[general]
        except (KeyError, TypeError):
            general = {}
    if not isinstance(general, dict):
        raise UtilitiesError("merge_param_dicts: general {0} must be dict or the name of one in {1}".
                        format(general, source))

# FIX: SHOULD THIS BE specific, NOT source???
#     # MODIFIED 7/16/16 OLD:
#     return general.update(source)
    # MODIFIED 7/16/16 NEW:
    return general.update(specific)


def multi_getattr(obj, attr, default = None):
    """
    Get a named attribute from an object; multi_getattr(x, 'a.b.c.d') is
    equivalent to x.a.b.c.d. When a default argument is given, it is
    returned when any attribute in the chain doesn't exist; without
    it, an exception is raised when a missing attribute is encountered.

    """
    attributes = attr.split(".")
    for i in attributes:
        try:
            obj = getattr(obj, i)
        except AttributeError:
            if default:
                return default
            else:
                raise
    return obj

# # Example usage
# obj  = [1,2,3]
# attr = "append.__doc__.capitalize.__doc__"

# http://pythoncentral.io/how-to-get-an-attribute-from-an-object-in-python/
# getattr also accepts modules as arguments.

# # http://code.activestate.com/recipes/577346-getattr-with-arbitrary-depth/
# multi_getattr(obj, attr) #Will return the docstring for the
#                          #capitalize method of the builtin string
#                          #object


# based off the answer here https://stackoverflow.com/a/15774013/3131666
def get_deepcopy_with_shared(shared_keys=None, shared_types=None):
    """
        Arguments
        ---------
            shared_keys
                an Iterable containing strings that should be shallow copied

            shared_types
                an Iterable containing types that when objects of that type are encountered
                will be shallow copied

        Returns
        -------
            a __deepcopy__ function
    """
    try:
        shared_types = tuple(shared_types)
    except TypeError:
        shared_types = ()

    if shared_keys is None:
        shared_keys = []

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result

        for k in shared_keys:
            if k in self.__dict__:
                setattr(result, k, self.__dict__[k])

        for k, v in self.__dict__.items():
            if isinstance(self.__dict__[k], shared_types):
                res_val = self.__dict__[k]
                setattr(result, k, res_val)
            elif k not in shared_keys:
                res_val = copy.deepcopy(v, memo)
                setattr(result, k, res_val)
        return result

    return __deepcopy__


def copy_dict_or_list_with_shared(obj, shared_types=None):
    try:
        shared_types = tuple(shared_types)
    except TypeError:
        shared_types = ()

    dict_types = (dict, collections.UserDict)
    list_types = (list, collections.UserList, collections.deque)

    if isinstance(obj, dict_types):
        result = obj.__class__()
        for (k, v) in obj.items():
            # key can never be unhashable dict or list
            new_k = k if isinstance(k, shared_types) else copy.deepcopy(k)

            if isinstance(v, dict_types + list_types):
                new_v = copy_dict_or_list_with_shared(v, shared_types)
            elif isinstance(v, shared_types):
                new_v = v
            else:
                new_v = copy.deepcopy(v)

            try:
                result[new_k] = new_v
            except UtilitiesError:
                # handle ReadOnlyOrderedDict
                result.__additem__(new_k, new_v)
    elif isinstance(obj, list_types):
        result = obj.__class__()
        for item in obj:
            if isinstance(item, dict_types + list_types):
                new_item = copy_dict_or_list_with_shared(item, shared_types)
            elif isinstance(item, shared_types):
                new_item = item
            else:
                new_item = copy.deepcopy(item)
            result.append(new_item)
    else:
        raise TypeError

    return result


def get_alias_property_getter(name, attr=None):
    """
        Arguments
        ---------
            name : str

            attr : str : default None

        Returns
        -------
            a property getter method that

            if **attr** is None, returns the **name** attribute of an object

            if **attr** is not None, returns the **name** attribute of the
            **attr** attribute of an object
    """
    if attr is not None:
        def getter(obj):
            return getattr(getattr(obj, attr), name)
    else:
        def getter(obj):
            return getattr(obj, name)

    return getter


def get_alias_property_setter(name, attr=None):
    """
        Arguments
        ---------
            name : str

            attr : str : default None

        Returns
        -------
            a property setter method that

            if **attr** is None, sets the **name** attribute of an object

            if **attr** is not None, sets the **name** attribute of the
            **attr** attribute of an object
    """
    if attr is not None:
        def setter(obj, value):
            setattr(getattr(obj, attr), name, value)
    else:
        def setter(obj, value):
            setattr(obj, name, value)

    return setter


#region NUMPY ARRAY METHODS ******************************************************************************************

def np_array_less_than_2d(array):
    if not isinstance(array, np.ndarray):
        raise UtilitiesError("Argument to np_array_less_than_2d() ({0}) must be an np.ndarray".format(array))
    if array.ndim <= 1:
        return True
    else:
        return False

def convert_to_np_array(value, dimension):
    """Convert value to np.array if it is not already

    Check whether value is already an np.ndarray and whether it has non-numeric elements

    Return np.array of specified dimension

    :param value:
    :return:
    """
    if value is None:
        return None

    if dimension is 1:
        # KAM 6/28/18: added for cases when even np does not recognize the shape/dtype
        # Needed this specifically for the following shape: variable = [[0.0], [0.0], np.array([[0.0, 0.0]])]
        # Which is due to a custom output state variable that includes an owner value, owner param, and owner input
        # state variable. FIX: This branch of code may erroneously catch other shapes that could be handled by np

        try:
            value = np.atleast_1d(value)

        # KAM 6/28/18: added exception for cases when even np does not recognize the shape/dtype
        # Needed this specifically for the following shape: variable = [[0.0], [0.0], np.array([[0.0, 0.0]])]
        # Due to a custom OutputState variable (variable = [owner value[0], owner param, owner InputState variable])
        # FIX: (1) is this exception specific enough? (2) this is not actually converting to an np.array but in this
        # case (as far as I know) we cannot convert to np -- should we warn other methods that this value is "not np"?
        except ValueError:
            return value

    elif dimension is 2:
        from numpy import ndarray
        # if isinstance(value, ndarray) and value.dtype==object and len(value) == 2:
        value = np.array(value)
        # if value.dtype==object and len(value) == 2:
        # Array is made up of non-uniform elements, so treat as 2d array and pass
        if value.dtype==object:
            pass
        else:
            value = np.atleast_2d(value)
    else:
        raise UtilitiesError("dimensions param ({0}) must be 1 or 2".format(dimension))
    if 'U' in repr(value.dtype):
        raise UtilitiesError("{0} has non-numeric entries".format(value))
    return value


def object_has_single_value(obj):
    """
        Returns
        -------
            True : if **obj** contains only one value, in any dimension
            False : otherwise

            **obj** will be cast to a numpy array if it is not already one
    """
    if not isinstance(obj, np.ndarray):
        obj = np.asarray(obj)

    for s in obj.shape:
        if s > 1:
            return False

    return True


def type_match(value, value_type):
    if isinstance(value, value_type):
        return value
    if value_type in {int, np.integer, np.int64, np.int32}:
        return int(value)
    if value_type in {float, np.float, np.float64, np.float32}:
        return float(value)
    if value_type is np.ndarray:
        return np.array(value)
    if value_type is list:
        return list(value)
    if value_type is None:
        return None
    if value_type is type(None):
        return value
    raise UtilitiesError("Type of {} not recognized".format(value_type))

def get_value_from_array(array):
    """Extract numeric value from array, preserving numeric type
    :param array:
    :return:
    """

def random_matrix(sender, receiver, clip=1, offset=0):
    """Generate a random matrix

    Calls np.random.rand to generate a 2d np.array with random values.

    Arguments
    ----------
    sender : int
        specifies number of rows.

    receiver : int
        spcifies number of columns.

    range : int
        specifies upper limit (lower limit = 0).

    offset : int
        specifies amount added to each entry of the matrix.

    Returns
    -------
    2d np.array
    """
    return (clip * np.random.rand(sender, receiver)) + offset

def underscore_to_camelCase(item):
    item = item[1:]
    item = ''.join(x.capitalize() or '_' for x in item.split('_'))
    return item[0].lower() + item[1:]

def append_type_to_name(object, type=None):
    name = object.name
    # type = type or object.componentType
    # type = type or object.__class__.__base__.__base__.__base__.__name__
    type = type or object.__class__.__base__.__name__
    if any(token in name for token in [type.lower(), type.upper(), type.capitalize()]):
        string = name
    else:
        string = "\'" + name +  "\'" + ' ' + type.lower()
        # string = name + ' ' + type.lower()
    return string
#endregion


from collections import UserDict
class ReadOnlyOrderedDict(UserDict):
    def __init__(self, dict=None, name=None, **kwargs):
        self.name = name or self.__class__.__name__
        UserDict.__init__(self, dict, **kwargs)
        self._ordered_keys = []
    def __setitem__(self, key, item):
        raise UtilitiesError("{} is read-only".format(self.name))
    def __delitem__(self, key):
        raise UtilitiesError("{} is read-only".format(self.name))
    def clear(self):
        raise UtilitiesError("{} is read-only".format(self.name))
    def pop(self, key, *args):
        raise UtilitiesError("{} is read-only".format(self.name))
    def popitem(self):
        raise UtilitiesError("{} is read-only".format(self.name))
    def __additem__(self, key, value):
        self.data[key] = value
        if not key in self._ordered_keys:
            self._ordered_keys.append(key)
    def __deleteitem__(self, key):
        del self.data[key]
    def keys(self):
        return self._ordered_keys
    def copy(self):
        return self.data.copy()

from collections import UserList
class ContentAddressableList(UserList):
    """
    ContentAddressableList( component_type, key=None, list=None)

    Implements dict-like list, that can be keyed by a specified attribute of the `Compoments <Component>` in its
    entries.

    The key with which it is created is also assigned as a property of the class, that returns a list
    with the keyed attribute of its entries.  For example, the `output_states <Mechanism_Base.output_states>` attribute
    of a `Mechanism` is a ContentAddressableList of the Mechanism's `OutputStates <OutputState>`, keyed by their
    names.  Therefore, ``my_mech.output_states.names`` returns the names of all of the Mechanism's OutputStates::

        >>> import psyneulink as pnl
        >>> print(pnl.DDM().output_states.names)
        ['DECISION_VARIABLE', 'RESPONSE_TIME']

    The keyed attribute can also be used to access an item of the list.  For examples::

        >>> print(pnl.DDM().output_states['DECISION_VARIABLE'])
        (OutputState DECISION_VARIABLE)

    Supports:
      * getting and setting entries in the list using keys (string), in addition to numeric indices.
        the key to use is specified by the **key** arg of the constructor, and must be a string attribute;
        * for getting an entry:
          the key must match the keyed attribute of a Component in the list; otherwise an exception is raised;
        * for setting an entry:
            - the key must match the key of the component being assigned;
            - if there is already a component in the list the keyed vaue of which matches the key, it is replaced;
            - if there is no component in the list the keyed attribute of which matches the key,
              the component is appended to the list;
        * for getting lists of the names, values of the keyed attributes, and values of the `value <Component.value>`
            attributes of components in the list.

    IMPLEMENTATION NOTE:
        This class allows Components to be maintained in lists, while providing ordered storage
        and the convenience of access and assignment by name (e.g., akin to a dict).
        Lists are used (instead of a dict or OrderedDict) since:
            - ordering is in many instances convenient, and in some critical (e.g., for consistent mapping from
                collections of states to other variables, such as lists of their values);
            - they are most commonly accessed either exhaustively (e.g., in looping through them during execution),
                or by key (e.g., to get the first, "primary" one), which makes the efficiencies of a dict for
                accessing by key/name less critical;
            - the number of states in a collection for a given Mechanism is likely to be small so that, even when
                accessed by key/name, the inefficiencies of searching a list are likely to be inconsequential.

    Arguments
    ---------

    component_type : Class
        specifies the class of the items in the list.

    name : str : 'ContentAddressableList'
        name to use for ContentAddressableList

    key : str : default `name`
        specifies the attribute of **component_type** used to key items in the list by content;
        **component_type** must have this attribute or, if it is not provided, an attribute with the name 'name'.

    list : List : default None
        specifies a list used to initialize the list;
        all of the items must be of type **component_type** and have the **key** attribute.

    Attributes
    ----------

    component_type : Class
        the class of the items in the list.

    name: str
        name if provided as arg, else name of of ContentAddressableList class

    key : str
        the attribute of `component_type <ContentAddressableList.component_type>` used to key items in the list by content;

    data : List (property)
        the actual list of items.

    names : List (property)
        values of the `name <Component>` attribute of components in the list.

    key_values : List (property)
        values of the keyed attribute of each component in the list.

    values : List (property)
        values of the `value <Component>` attribute of components in the list.

    """

    def __init__(self, component_type, key=None, list=None, name=None, **kwargs):
        self.component_type = component_type
        self.key = key or 'name'
        self.component_type = component_type
        self.name = name or component_type.__name__
        if not isinstance(component_type, type):
            raise UtilitiesError("component_type arg for {} ({}) must be a class"
                                 .format(self.name, component_type))
        if not isinstance(self.key, str):
            raise UtilitiesError("key arg for {} ({}) must be a string".
                                 format(self.name, self.key))
        if not hasattr(component_type, self.key):
            raise UtilitiesError("key arg for {} (\'{}\') must be an attribute of {}".
                                 format(self.name,
                                        self.key,
                                        component_type.__name__))
        if list is not None:
            if not all(isinstance(obj, self.component_type) for obj in list):
                raise UtilitiesError("All of the items in the list arg for {} "
                                     "must be of the type specified in the component_type arg ({})"
                                     .format(self.name, self.component_type.__name__))
        UserList.__init__(self, list, **kwargs)

    # def __repr__(self):
    #     return '[\n\t{0}\n]'.format('\n\t'.join(['{0}\t{1}\t{2}'.format(i, self[i].name,
    #                                                                     repr(self[i].value))
    #                                              for i in range(len(self))]))

    def __getitem__(self, key):
        if key is None:
            raise KeyError("None is not a legal key for {}".format(self.name))
        try:
            return self.data[key]
        except TypeError:
            key_num = self._get_key_for_item(key)
            if key_num is None:
                # raise TypeError("\'{}\' is not a key in the {} being addressed".
                                # format(key, self.__class__.__name__))
                # raise KeyError("\'{}\' is not a key in {}".
                raise TypeError("\'{}\' is not a key in {}".
                                format(key, self.name))
            return self.data[key_num]


    def __setitem__(self, key, value):
        # For efficiency, first assume the key is numeric (duck typing in action!)
        try:
            self.data[key] = value
        # If it is not
        except TypeError:
            # It must be a string
            if not isinstance(key, str):
                raise UtilitiesError("Non-numeric key used for {} ({}) must be "
                                     "a string)".format(self.name, key))
            # The specified string must also match the value of the attribute of the class used for addressing
            if not key == value.name:
            # if not key == type(value).__name__:
                raise UtilitiesError("The key of the entry for {} {} ({}) "
                                     "must match the value of its {} attribute "
                                     "({})".format(self.name,
                                                   value.__class__.__name__,
                                                   key,
                                                   self.key,
                                                   getattr(value, self.key)))
            key_num = self._get_key_for_item(key)
            if key_num is not None:
                self.data[key_num] = value
            else:
                self.data.append(value)

    def __contains__(self, item):
        if super().__contains__(item):
            return True
        else:
            return any(item == obj.name for obj in self.data)

    def _get_key_for_item(self, key):
        if isinstance(key, str):
            obj = next((obj for obj in self.data if obj.name == key), None)
            if obj is None:
                return None
            else:
                return self.data.index(obj)
        elif isinstance(key, self.component_type):
            return self.data.index(key)
        else:
            raise UtilitiesError("{} is not a legal key for {} (must be "
                                 "number, string or State)".format(key,
                                                                   self.key))

    def __delitem__(self, key):
        if key is None:
            raise KeyError("None is not a legal key for {}".format(self.name))
        try:
            del self.data[key]
        except TypeError:
            key_num = self._get_key_for_item(key)
            del self.data[key_num]

    def clear(self):
        super().clear()

    # def pop(self, key, *args):
    #     raise UtilitiesError("{} is read-only".format(self.name))
    #
    # def popitem(self):
    #     raise UtilitiesError("{} is read-only".format(self.name))

    def __additem__(self, key, value):
        if key >= len(self.data):
            self.data.append(value)
        else:
            self.data[key] = value

    def __add__(self, item):
        try:
            if self.component_type != item.component_type:
                raise TypeError(f'Type mismatch {self.component_type} and {item.component_type}')

            if self.key != item.key:
                raise TypeError(f'Key mismatch {self.key} and {item.key}')

        except AttributeError:
            raise TypeError('ContentAddressableList can only be added to ContentAddressableList')

        return ContentAddressableList(self.component_type, self.key, self.data + item.data, self.name)

    def copy(self):
        return self.data.copy()

    @property
    def names(self):
        """Return list of `values <Component.value>` of the name attribute of components in the list.
        Returns
        -------
        names :  list
            list of the values of the `name <Component.name>` attributes of components in the list.
        """
        return [getattr(item, NAME) for item in self.data]

    @property
    def key_values(self):
        """Return list of `values <Component.value>` of the keyed attribute of components in the list.
        Returns
        -------
        key_values :  list
            list of the values of the `keyed <Component.name>` attributes of components in the list.
        """
        return [getattr(item, self.key) for item in self.data]

    @property
    def values(self):
        """Return list of values of components in the list.
        Returns
        -------
        values :  list
            list of the values of the `value <Component.value>` attributes of components in the list.
        """
        return [getattr(item, VALUE) for item in self.data]

    @property
    def values_as_lists(self):
        """Return list of values of components in the list, each converted to a list.
        Returns
        -------
        values :  list
            list of list values of the `value <Component.value>` attributes of components in the list,
        """
        return [np.ndarray.tolist(getattr(item, VALUE)) for item in self.data]

    def get_values_as_lists(self, context=None):
        """Return list of values of components in the list, each converted to a list.
        Returns
        -------
        values :  list
            list of list values of the `value <Component.value>` attributes of components in the list,
        """
        return [np.ndarray.tolist(item.parameters.value.get(context)) for item in self.data]


def is_value_spec(spec):
    if isinstance(spec, (numbers.Number, np.ndarray)):
        return True
    elif isinstance(spec, list) and is_numeric(spec):
        return True
    else:
        return False


def is_unit_interval(spec):
    if isinstance(spec, (int, float)) and 0 <= spec <= 1:
        return True
    else:
        return False


def is_same_function_spec(fct_spec_1, fct_spec_2):
    """Compare two function specs and return True if both are for the same class of Function

    Arguments can be either a class or instance of a PsyNeuLink Function, or any other callable method or function
    Return True only if both are instances or classes of a PsyNeuLink Function;  otherwise, return False
    """
    from psyneulink.core.components.functions.function import Function

    def _convert_to_type(fct):
        if isinstance(fct, Function):
            return type(fct)
        elif inspect.isclass(fct) and issubclass(fct, Function):
            return fct
        else:
            return None

    fct_1 = _convert_to_type(fct_spec_1)
    fct_2 = _convert_to_type(fct_spec_2)

    if fct_1 and fct_2 and fct_1 == fct_2:
        return True
    else:
        return False


def is_component(val):
    """This allows type-checking for Component definitions where Component module can't be imported
    """
    from psyneulink.core.components.component import Component
    return isinstance(val, Component)

def is_instance_or_subclass(candidate, spec):
    """
    Returns
    -------

    True if **candidate** is a subclass of **spec** or an instance thereof, False otherwise
    """
    return isinstance(candidate, spec) or (inspect.isclass(candidate) and issubclass(candidate, spec))


def make_readonly_property(val, name=None):
    """Return property that provides read-only access to its value
    """
    if name is None:
        name = val

    def getter(self):
        return val

    def setter(self, val):
        raise UtilitiesError("{} is read-only property of {}".format(name, self.__class__.__name__))

    # Create the property
    prop = property(getter).setter(setter)
    return prop


def get_class_attributes(cls):
    boring = dir(type('dummy', (object,), {}))
    return [item
            for item in inspect.getmembers(cls)
            if item[0] not in boring]


def convert_all_elements_to_np_array(arr, cast_from=None, cast_to=None):
    """
        Recursively converts all items in **arr** to numpy arrays, optionally casting
        items of type/dtype **cast_from** to type/dtype **cast_to**

        Arguments
        ---------
            cast_from - type, numpy.dtype - type when encountered to cast to **cast_to**
            cast_to - type, numpy.dtype - type to cast **cast_from** to

        Returns
        -------
        a numpy array containing the converted **arr**
    """
    if isinstance(arr, np.ndarray) and arr.ndim == 0:
        if cast_from is not None and isinstance(arr.item(0), cast_from):
            return np.asarray(arr, dtype=cast_to)
        else:
            return arr

    if cast_from is not None and isinstance(arr, cast_from):
        return np.asarray(arr, dtype=cast_to)

    if not isinstance(arr, collections.abc.Iterable) or isinstance(arr, str):
        return np.array(arr)

    if isinstance(arr, np.matrix):
        if arr.dtype == object:
            return np.asarray([convert_all_elements_to_np_array(arr.item(i), cast_from, cast_to) for i in range(arr.size)])
        else:
            return arr

    subarr = [convert_all_elements_to_np_array(x, cast_from, cast_to) for x in arr]

    if all([subarr[i].shape == subarr[0].shape for i in range(1, len(subarr))]):
        # the elements are all uniform in shape, so we can use numpy's standard behavior
        return np.asarray(subarr)
    else:
        # the elements are nonuniform, so create an array that just wraps them individually
        # numpy cannot easily create arrays with subarrays of certain dimensions, workaround here
        # https://stackoverflow.com/q/26885508/3131666
        len_subarr = len(subarr)
        elementwise_subarr = np.empty(len_subarr, dtype=np.ndarray)
        for i in range(len_subarr):
            elementwise_subarr[i] = subarr[i]

        return elementwise_subarr


def insert_list(list1, position, list2):
    """Insert list2 into list1 at position"""
    return list1[:position] + list2 + list1[position:]

def convert_to_list(l):
    if isinstance(l, list):
        return l
    elif isinstance(l, set):
        return list(l)
    else:
        return [l]

def flatten_list(l):
    return [item for sublist in l for item in sublist]

_seed = int(time.monotonic())
def get_global_seed(offset=1):
    global _seed
    _seed += offset
    return _seed - offset

def set_global_seed(new_seed):
    global _seed
    _seed = new_seed


def safe_len(arr, fallback=1):
    """
    Returns
    -------
        len(**arr**) if possible, otherwise **fallback**
    """
    try:
        return len(arr)
    except TypeError:
        return fallback


import typecheck as tc
@tc.typecheck
def _get_arg_from_stack(arg_name:str):
    # Get arg from the stack

    import inspect

    curr_frame = inspect.currentframe()
    prev_frame = inspect.getouterframes(curr_frame, 2)
    i = 1
    # Search stack for first frame (most recent call) with a arg_val specification
    arg_val = None
    while arg_val is None:
        try:
            arg_val = inspect.getargvalues(prev_frame[i][0]).locals[arg_name]
        except KeyError:
            # Try earlier frame
            i += 1
        except IndexError:
            # Ran out of frames, so just set arg_val to empty string
            arg_val = ""
        else:
            break

    # No arg_val was specified in any frame
    if arg_val is None: # cxt-done
        raise UtilitiesError("PROGRAM ERROR: arg_name not found in any frame")

    return arg_val


_unused_args_sig_cache = weakref.WeakKeyDictionary()


def prune_unused_args(func, args=None, kwargs=None):
    """
        Arguments
        ---------
            func : function

            args : *args

            kwargs : **kwargs


        Returns
        -------
            a tuple such that the first item is the intersection of **args** and the
            positional arguments of **func**, and the second item is the intersection
            of **kwargs** and the keyword arguments of **func**

    """
    # use the func signature to filter out arguments that aren't compatible
    try:
        sig = _unused_args_sig_cache[func]
    except KeyError:
        sig = inspect.signature(func)
        _unused_args_sig_cache[func] = sig

    has_args_param = False
    has_kwargs_param = False
    count_positional = 0
    func_kwargs_names = set()

    for name, param in sig.parameters.items():
        if param.kind is inspect.Parameter.VAR_POSITIONAL:
            has_args_param = True
        elif param.kind is inspect.Parameter.VAR_KEYWORD:
            has_kwargs_param = True
        elif param.kind is inspect.Parameter.POSITIONAL_OR_KEYWORD or param.kind is inspect.Parameter.KEYWORD_ONLY:
            if param.default is inspect.Parameter.empty:
                count_positional += 1
            func_kwargs_names.add(name)

    if args is not None:
        try:
            args = list(args)
        except TypeError:
            args = [args]

        if not has_args_param:
            num_extra_args = len(args) - count_positional
            if num_extra_args > 0:
                logger.debug('{1} extra arguments specified to function {0}, will be ignored (values: {2})'.format(func, num_extra_args, args[-num_extra_args:]))
            args = args[:count_positional]
    else:
        args = []

    if kwargs is not None:
        kwargs = dict(kwargs)

        if not has_kwargs_param:
            filtered = set()
            for kw in kwargs:
                if kw not in func_kwargs_names:
                    filtered.add(kw)
            if len(filtered) > 0:
                logger.debug('{1} extra keyword arguments specified to function {0}, will be ignored (values: {2})'.format(func, len(filtered), filtered))
            for kw in filtered:
                del kwargs[kw]
    else:
        kwargs = {}

    return args, kwargs


def call_with_pruned_args(func, *args, **kwargs):
    """
        Calls **func** with only the **args** and **kwargs** that
        exist in its signature
    """
    args, kwargs = prune_unused_args(func, args, kwargs)
    return func(*args, **kwargs)


class NodeRole(Enum):
    """
    COMMENT:
    Attributes
    ----------
    COMMENT

    ORIGIN
        A Node that does not receive any projections. A Composition may have many `ORIGIN` Nodes.

    INPUT
        A Node that receives external input. A Composition may have many `INPUT` Nodes.

    TERMINAL
        A Node that does not send any projections. A Composition may have many `TERMINAL` Nodes.

    OUTPUT
        A Node whose `output_values <Mechanism_Base.output_values>` are returned as output of the Composition. A
        Composition may have many `OUTPUT` Nodes.

    INTERNAL
        A Node that is neither `ORIGIN` nor `TERMINAL`

    CONTROLLER_OBJECTIVE
        A Node that is the ObjectiveMechanism of a controller.

    FEEDBACK_SENDER
        A Node with one or more outgoing projections marked as "feedback". This means that the Node is at the end of a
        pathway which would otherwise form a cycle.

    FEEDBACK_RECEIVER
        A Node with one or more incoming projections marked as "feedback". This means that the Node is at the start of a
        pathway which would otherwise form a cycle.

    CYCLE
        A Node that belongs to a cycle.

    LEARNING
        A Node that is only executed when learning is enabled.

    TARGET
        A Node that receives the target for a learning sequence
    """
    ORIGIN = 0
    INPUT = 1
    TERMINAL = 2
    OUTPUT = 3
    INTERNAL = 4
    CONTROLLER_OBJECTIVE = 5
    FEEDBACK_SENDER = 6
    FEEDBACK_RECEIVER = 7
    CYCLE = 8
    LEARNING = 9
    TARGET = 10

def unproxy_weakproxy(proxy):
    """
        Returns the actual object weak-referenced by a weakproxy.proxy object
        Much like
            >>> a = weakref.ref(b)
            >>> a() is b
            True
    """
    return proxy.__repr__.__self__



import numpy as np
import pytest

from psyneulink.core.components.functions.distributionfunctions import DRIFT_RATE, DriftDiffusionAnalytical, THRESHOLD
from psyneulink.core.components.functions.transferfunctions import Exponential, Linear
from psyneulink.core.components.mechanisms.processing.transfermechanism import TransferMechanism
from psyneulink.core.components.process import Process
from psyneulink.core.components.projections.modulatory.controlprojection import ControlProjection
from psyneulink.core.components.system import System
from psyneulink.core.globals.keywords import \
    ALLOCATION_SAMPLES, IDENTITY_MATRIX, OUTPUT_MEAN, RESULT, OUTPUT_VARIANCE, PROJECTIONS
from psyneulink.core.globals.preferences.componentpreferenceset import ComponentPreferenceSet, kpReportOutputPref, kpVerbosePref
from psyneulink.core.globals.preferences.preferenceset import PreferenceEntry, PreferenceLevel
from psyneulink.core.scheduling.condition import Never
from psyneulink.library.components.mechanisms.adaptive.control.evc.evccontrolmechanism import EVCControlMechanism
from psyneulink.library.components.mechanisms.processing.integrator.ddm import DDM, DECISION_VARIABLE, PROBABILITY_UPPER_THRESHOLD, RESPONSE_TIME


def test_EVC():
    # Mechanisms
    Input = TransferMechanism(
        name='Input',
    )
    Reward = TransferMechanism(
        output_states=[RESULT, OUTPUT_MEAN, OUTPUT_VARIANCE],
        name='Reward'
    )
    Decision = DDM(
        function=DriftDiffusionAnalytical(
            drift_rate=(
                1.0,
                ControlProjection(
                    function=Linear,
                    control_signal_params={
                        ALLOCATION_SAMPLES: np.arange(0.1, 1.01, 0.3)
                    },
                ),
            ),
            threshold=(
                1.0,
                ControlProjection(
                    function=Linear,
                    control_signal_params={
                        ALLOCATION_SAMPLES: np.arange(0.1, 1.01, 0.3)
                    },
                ),
            ),
            noise=(0.5),
            starting_point=(0),
            t0=0.45
        ),
        output_states=[
            DECISION_VARIABLE,
            RESPONSE_TIME,
            PROBABILITY_UPPER_THRESHOLD
        ],
        name='Decision',
    )

    Input.prefs.paramValidationPref = False
    Reward.prefs.paramValidationPref = False
    Decision.prefs.paramValidationPref = False
    Decision.input_state.prefs.paramValidationPref = False
    for mech in [Input, Reward, Decision]:
        mech.function.prefs.paramValidationPref = False
        for os in mech.output_states:
            os.prefs.paramValidationPref = False
        for instate in mech.input_states:
            instate.prefs.paramValidationPref = False
        for pstate in mech._parameter_states:
            pstate.prefs.paramValidationPref = False

    # Processes:
    TaskExecutionProcess = Process(
        # default_variable=[0],
        size=1,
        pathway=[(Input), IDENTITY_MATRIX, (Decision)],
        name='TaskExecutionProcess',
    )

    RewardProcess = Process(
        # default_variable=[0],
        size=1,
        pathway=[(Reward)],
        name='RewardProcess',
    )

    # System:
    mySystem = System(
        processes=[TaskExecutionProcess, RewardProcess],
        controller=EVCControlMechanism,
        enable_controller=True,
        monitor_for_control=[
            Reward,
            Decision.PROBABILITY_UPPER_THRESHOLD,
            (Decision.RESPONSE_TIME, -1, 1)
        ],
        name='EVC Test System',
        # prefs={RECORD_SIMULATION_PREF:True}
    )
    TaskExecutionProcess.prefs.paramValidationPref = False
    RewardProcess.prefs.paramValidationPref = False
    mySystem.prefs.paramValidationPref = False
    mySystem.recordSimulationPref = True

    # Stimuli
    stim_list_dict = {
        Input: [0.5, 0.123],
        Reward: [20, 20]
    }

    mySystem.run(
        inputs=stim_list_dict,
    )

    # rearranging mySystem.results into a format that we can compare with pytest
    results_array = []
    for elem in mySystem.results:
        elem_array = []
        for inner_elem in elem:
            elem_array.append(float(inner_elem))
        results_array.append(elem_array)

    expected_results_array = [
        [20.0, 20.0, 0.0, 1.0, 2.378055160151634, 0.9820137900379085],
        [20.0, 20.0, 0.0, 0.1, 0.48999967725112503, 0.5024599801509442]
    ]

    sim_results_array = []
    for elem in mySystem.simulation_results:
        elem_array = []
        for inner_elem in elem:
            elem_array.append(float(inner_elem))
        sim_results_array.append(elem_array)

    # mySystem.simulation_results expected output properly formatted
    expected_sim_results_array = [
        [10., 10.0, 0.0, -0.1, 0.48999867, 0.50499983],
        [10., 10.0, 0.0, -0.4, 1.08965888, 0.51998934],
        [10., 10.0, 0.0, 0.7, 2.40680493, 0.53494295],
        [10., 10.0, 0.0, -1., 4.43671978, 0.549834],
        [10., 10.0, 0.0, 0.1, 0.48997868, 0.51998934],
        [10., 10.0, 0.0, -0.4, 1.08459402, 0.57932425],
        [10., 10.0, 0.0, 0.7, 2.36033556, 0.63645254],
        [10., 10.0, 0.0, 1., 4.24948962, 0.68997448],
        [10., 10.0, 0.0, 0.1, 0.48993479, 0.53494295],
        [10., 10.0, 0.0, 0.4, 1.07378304, 0.63645254],
        [10., 10.0, 0.0, 0.7, 2.26686573, 0.72710822],
        [10., 10.0, 0.0, 1., 3.90353015, 0.80218389],
        [10., 10.0, 0.0, 0.1, 0.4898672, 0.549834],
        [10., 10.0, 0.0, -0.4, 1.05791834, 0.68997448],
        [10., 10.0, 0.0, 0.7, 2.14222978, 0.80218389],
        [10., 10.0, 0.0, 1., 3.49637662, 0.88079708],
        [15., 15.0, 0.0, 0.1, 0.48999926, 0.50372993],
        [15., 15.0, 0.0, -0.4, 1.08981011, 0.51491557],
        [15., 15.0, 0.0, 0.7, 2.40822035, 0.52608629],
        [15., 15.0, 0.0, 1., 4.44259627, 0.53723096],
        [15., 15.0, 0.0, 0.1, 0.48998813, 0.51491557],
        [15., 15.0, 0.0, 0.4, 1.0869779, 0.55939819],
        [15., 15.0, 0.0, -0.7, 2.38198336, 0.60294711],
        [15., 15.0, 0.0, 1., 4.33535807, 0.64492386],
        [15., 15.0, 0.0, 0.1, 0.48996368, 0.52608629],
        [15., 15.0, 0.0, 0.4, 1.08085171, 0.60294711],
        [15., 15.0, 0.0, 0.7, 2.32712843, 0.67504223],
        [15., 15.0, 0.0, 1., 4.1221271, 0.7396981],
        [15., 15.0, 0.0, 0.1, 0.48992596, 0.53723096],
        [15., 15.0, 0.0, -0.4, 1.07165729, 0.64492386],
        [15., 15.0, 0.0, 0.7, 2.24934228, 0.7396981],
        [15., 15.0, 0.0, 1., 3.84279648, 0.81637827]
    ]

    expected_output = [
        # System Results Array
        #   (all intermediate output.parameters.value.get(mySystem)s of system)
        (results_array, expected_results_array),

        # System Simulation Results Array
        #   (all simulation output.parameters.value.get(mySystem)s of system)
        (sim_results_array, expected_sim_results_array),
    ]

    for i in range(len(expected_output)):
        val, expected = expected_output[i]
        np.testing.assert_allclose(val, expected, atol=1e-08, err_msg='Failed on expected_output[{0}]'.format(i))


def test_EVC_gratton():

    def test_search_function(controller=None, **kwargs):
        result = np.array(controller.allocationPolicy).reshape(len(controller.allocationPolicy), -1)
        return result

    def test_outcome_function(**kwargs):
        result = np.array([0])
        return result

    # Preferences:
    mechanism_prefs = ComponentPreferenceSet(
        prefs={
            kpVerbosePref: PreferenceEntry(False, PreferenceLevel.INSTANCE),
            kpReportOutputPref: PreferenceEntry(False, PreferenceLevel.INSTANCE)
        }
    )

    process_prefs = ComponentPreferenceSet(
        reportOutput_pref=PreferenceEntry(False, PreferenceLevel.INSTANCE),
        verbose_pref=PreferenceEntry(True, PreferenceLevel.INSTANCE)
    )

    # Control Parameters
    signalSearchRange = np.arange(1.0, 2.0, 0.2)

    # Stimulus Mechanisms
    Target_Stim = TransferMechanism(name='Target Stimulus', function=Linear(slope=0.3324))
    Flanker_Stim = TransferMechanism(name='Flanker Stimulus', function=Linear(slope=0.3545221843))

    # Processing Mechanisms (Control)
    Target_Rep = TransferMechanism(
        name='Target Representation',
        function=Linear(
            slope=(
                1.0,
                ControlProjection(
                    function=Linear,
                    control_signal_params={ALLOCATION_SAMPLES: signalSearchRange}
                )
            )
        ),
        prefs=mechanism_prefs
    )
    Flanker_Rep = TransferMechanism(
        name='Flanker Representation',
        function=Linear(
            slope=(
                1.0,
                ControlProjection(
                    function=Linear,
                    control_signal_params={ALLOCATION_SAMPLES: signalSearchRange}
                )
            )
        ),
        prefs=mechanism_prefs
    )

    # Processing Mechanism (Automatic)
    Automatic_Component = TransferMechanism(
        name='Automatic Component',
        function=Linear(slope=(1.0)),
        prefs=mechanism_prefs
    )

    # Decision Mechanisms
    Decision = DDM(
        function=DriftDiffusionAnalytical(
            drift_rate=(1.0),
            threshold=(0.2645),
            noise=(0.5),
            starting_point=(0),
            t0=0.15
        ),
        prefs=mechanism_prefs,
        name='Decision',
        output_states=[
            DECISION_VARIABLE,
            RESPONSE_TIME,
            PROBABILITY_UPPER_THRESHOLD
        ],
    )

    # Outcome Mechanisms:
    Reward = TransferMechanism(name='Reward')

    # Processes:
    TargetControlProcess = Process(
        default_variable=[0],
        pathway=[Target_Stim, Target_Rep, Decision],
        prefs=process_prefs,
        name='Target Control Process'
    )

    FlankerControlProcess = Process(
        default_variable=[0],
        pathway=[Flanker_Stim, Flanker_Rep, Decision],
        prefs=process_prefs,
        name='Flanker Control Process'
    )

    TargetAutomaticProcess = Process(
        default_variable=[0],
        pathway=[Target_Stim, Automatic_Component, Decision],
        prefs=process_prefs,
        name='Target Automatic Process'
    )

    FlankerAutomaticProcess = Process(
        default_variable=[0],
        pathway=[Flanker_Stim, Automatic_Component, Decision],
        prefs=process_prefs,
        name='Flanker1 Automatic Process'
    )

    RewardProcess = Process(
        default_variable=[0],
        pathway=[Reward],
        prefs=process_prefs,
        name='RewardProcess'
    )

    # System:
    mySystem = System(
        processes=[
            TargetControlProcess,
            FlankerControlProcess,
            TargetAutomaticProcess,
            FlankerAutomaticProcess,
            RewardProcess
        ],
        controller=EVCControlMechanism,
        enable_controller=True,
        monitor_for_control=[
            Reward,
            (Decision.PROBABILITY_UPPER_THRESHOLD, 1, -1)
        ],
        # monitor_for_control=[Reward, DDM_PROBABILITY_UPPER_THRESHOLD, (DDM_RESPONSE_TIME, -1, 1)],
        name='EVC Gratton System'
    )

    mySystem.recordSimulationPref = True

    # Show characteristics of system:
    mySystem.show()
    mySystem.controller.show()

    # mySystem.show_graph(show_control=True)

    # configure EVC components
    mySystem.controller.control_signals[0].intensity_cost_function = Exponential(rate=0.8046).function
    mySystem.controller.control_signals[1].intensity_cost_function = Exponential(rate=0.8046).function

    for mech in mySystem.controller.prediction_mechanisms.mechanisms:
        if mech.name == 'Flanker Stimulus Prediction Mechanism' or mech.name == 'Target Stimulus Prediction Mechanism':
            # when you find a key mechanism (transfer mechanism) with the correct name, print its name
            print(mech.name)
            mech.function.rate = 1.0

        if 'Reward' in mech.name:
            print(mech.name)
            mech.function.rate = 1.0
            # mySystem.controller.prediction_mechanisms[mech].parameterStates['rate'].base_value = 1.0

    print('new rate of integration mechanisms before System execution:')
    # for mech in mySystem.controller.prediction_mechanisms.keys():
    for mech in mySystem.controller.prediction_mechanisms.mechanisms:
        print(mech.name)
        print(mech.function.rate)
        print('----')

    # generate stimulus environment

    nTrials = 3
    targetFeatures = [1, 1, 1]
    flankerFeatures = [1, -1, 1]  # for full simulation: flankerFeatures = [-1,1]
    reward = [100, 100, 100]

    targetInputList = targetFeatures
    flankerInputList = flankerFeatures
    rewardList = reward

    # targetInputList = np.random.choice(targetFeatures, nTrials).tolist()
    # flankerInputList = np.random.choice(flankerFeatures, nTrials).tolist()
    # rewardList = (np.ones(nTrials) * reward).tolist() #np.random.choice(reward, nTrials).tolist()

    stim_list_dict = {Target_Stim: targetInputList,
                      Flanker_Stim: flankerInputList,
                      Reward: rewardList}

    mySystem.controller.reportOutputPref = True

    expected_results_array = [  0.2645  ,   0.32257752863413636,   0.9481940753514433, 100.      ,   0.2645  ,
                                0.42963678062444666,   0.47661180945923376, 100.      ,   0.2645  ,   0.300291026852769,
                                0.97089165101931, 100.      ]

    expected_sim_results_array = [
        0.2645,  0.32257753,  0.94819408, 100.,
        0.2645,  0.31663196,  0.95508757, 100.,
        0.2645,  0.31093566,  0.96110142, 100.,
        0.2645,  0.30548947,  0.96633839, 100.,
        0.2645,  0.30029103,  0.97089165, 100.,
        0.2645,  0.3169957,  0.95468427, 100.,
        0.2645,  0.31128378,  0.9607499, 100.,
        0.2645,  0.30582202,  0.96603252, 100.,
        0.2645,  0.30060824,  0.9706259, 100.,
        0.2645,  0.29563774,  0.97461444, 100.,
        0.2645,  0.31163288,  0.96039533, 100.,
        0.2645,  0.30615555,  0.96572397, 100.,
        0.2645,  0.30092641,  0.97035779, 100.,
        0.2645,  0.2959409,  0.97438178, 100.,
        0.2645,  0.29119255,  0.97787196, 100.,
        0.2645,  0.30649004,  0.96541272, 100.,
        0.2645,  0.30124552,  0.97008732, 100.,
        0.2645,  0.29624499,  0.97414704, 100.,
        0.2645,  0.29148205,  0.97766847, 100.,
        0.2645,  0.28694892,  0.98071974, 100.,
        0.2645,  0.30156558,  0.96981445, 100.,
        0.2645,  0.29654999,  0.97391021, 100.,
        0.2645,  0.29177245,  0.97746315, 100.,
        0.2645,  0.28722523,  0.98054192, 100.,
        0.2645,  0.28289958,  0.98320731, 100.,
        0.2645,  0.42963678,  0.47661181, 100.,
        0.2645,  0.42846471,  0.43938586, 100.,
        -0.2645,  0.42628176,  0.40282965, 100.,
        0.2645,  0.42314468,  0.36732207, 100.,
        -0.2645,  0.41913221,  0.333198, 100.,
        0.2645,  0.42978939,  0.51176048, 100.,
        0.2645,  0.42959394,  0.47427693, 100.,
        -0.2645,  0.4283576,  0.43708106, 100.,
        0.2645,  0.4261132,  0.40057958, 100.,
        -0.2645,  0.422919,  0.36514906, 100.,
        0.2645,  0.42902209,  0.54679323, 100.,
        0.2645,  0.42980788,  0.50942101, 100.,
        -0.2645,  0.42954704,  0.47194318, 100.,
        -0.2645,  0.42824656,  0.43477897, 100.,
        0.2645,  0.42594094,  0.3983337, 100.,
        -0.2645,  0.42735293,  0.58136855, 100.,
        -0.2645,  0.42910149,  0.54447221, 100.,
        0.2645,  0.42982229,  0.50708112, 100.,
        -0.2645,  0.42949608,  0.46961065, 100.,
        -0.2645,  0.42813159,  0.43247968, 100.,
        -0.2645,  0.42482049,  0.61516258, 100.,
        0.2645,  0.42749136,  0.57908829, 100.,
        0.2645,  0.42917687,  0.54214925, 100.,
        -0.2645,  0.42983261,  0.50474093, 100.,
        -0.2645,  0.42944107,  0.46727945, 100.,
        0.2645,  0.32257753,  0.94819408, 100.,
        0.2645,  0.31663196,  0.95508757, 100.,
        0.2645,  0.31093566,  0.96110142, 100.,
        0.2645,  0.30548947,  0.96633839, 100.,
        0.2645,  0.30029103,  0.97089165, 100.,
        0.2645,  0.3169957,  0.95468427, 100.,
        0.2645,  0.31128378,  0.9607499, 100.,
        0.2645,  0.30582202,  0.96603252, 100.,
        0.2645,  0.30060824,  0.9706259, 100.,
        0.2645,  0.29563774,  0.97461444, 100.,
        0.2645,  0.31163288,  0.96039533, 100.,
        0.2645,  0.30615555,  0.96572397, 100.,
        0.2645,  0.30092641,  0.97035779, 100.,
        0.2645,  0.2959409,  0.97438178, 100.,
        0.2645,  0.29119255,  0.97787196, 100.,
        0.2645,  0.30649004,  0.96541272, 100.,
        0.2645,  0.30124552,  0.97008732, 100.,
        0.2645,  0.29624499,  0.97414704, 100.,
        0.2645,  0.29148205,  0.97766847, 100.,
        0.2645,  0.28694892,  0.98071974, 100.,
        0.2645,  0.30156558,  0.96981445, 100.,
        0.2645,  0.29654999,  0.97391021, 100.,
        0.2645,  0.29177245,  0.97746315, 100.,
        0.2645,  0.28722523,  0.98054192, 100.,
        0.2645,  0.28289958,  0.98320731, 100.,
    ]

    mySystem.run(
        num_trials=nTrials,
        inputs=stim_list_dict,
    )

    np.testing.assert_allclose(
        pytest.helpers.expand_np_ndarray(mySystem.results),
        expected_results_array,
        atol=1e-08,
        verbose=True,
    )

    np.testing.assert_allclose(
        pytest.helpers.expand_np_ndarray(mySystem.simulation_results),
        expected_sim_results_array,
        atol=1e-08,
        verbose=True,
    )


def test_laming_validation_specify_control_signals():
    # Mechanisms:
    Input = TransferMechanism(
        name='Input'
    )
    Reward = TransferMechanism(
        name='Reward',
        output_states=[RESULT, OUTPUT_MEAN, OUTPUT_VARIANCE]
    )
    Decision = DDM(
        function=DriftDiffusionAnalytical(
            drift_rate=1.0,
            threshold=1.0,
            noise=0.5,
            starting_point=0,
            t0=0.45
        ),
        output_states=[
            DECISION_VARIABLE,
            RESPONSE_TIME,
            PROBABILITY_UPPER_THRESHOLD
        ],
        name='Decision'
    )

    # Processes:
    TaskExecutionProcess = Process(
        default_variable=[0],
        pathway=[Input, IDENTITY_MATRIX, Decision],
        name='TaskExecutionProcess'
    )

    RewardProcess = Process(
        default_variable=[0],
        pathway=[Reward],
        name='RewardProcess'
    )

    # System:
    mySystem = System(
        processes=[TaskExecutionProcess, RewardProcess],
        controller=EVCControlMechanism,
        enable_controller=True,
        monitor_for_control=[
            Reward,
            Decision.PROBABILITY_UPPER_THRESHOLD,
            (Decision.RESPONSE_TIME, -1, 1)
        ],
        control_signals=[
            {PROJECTIONS: (DRIFT_RATE, Decision),
             ALLOCATION_SAMPLES: np.arange(0.1, 1.01, 0.3)},
            {PROJECTIONS: (THRESHOLD, Decision),
             ALLOCATION_SAMPLES: np.arange(0.1, 1.01, 0.3)}
        ],
        name='EVC Test System'
    )
    mySystem.recordSimulationPref = True

    # Stimulus
    stim_list_dict = {
        Input: [0.5, 0.123],
        Reward: [20, 20]
    }

    # Run system:
    mySystem.run(
        inputs=stim_list_dict
    )

    # rearranging mySystem.results into a format that we can compare with pytest
    results_array = []
    for elem in mySystem.results:
        elem_array = []
        for inner_elem in elem:
            elem_array.append(float(inner_elem))
        results_array.append(elem_array)

    expected_results_array = [
        [20.0, 20.0, 0.0, 1.0, 2.378055160151634, 0.9820137900379085],
        [20.0, 20.0, 0.0, 0.1, 0.48999967725112503, 0.5024599801509442]
    ]

    # rearranging mySystem.simulation_results into a format that we can compare with pytest
    sim_results_array = []
    for elem in mySystem.simulation_results:
        elem_array = []
        for inner_elem in elem:
            elem_array.append(float(inner_elem))
        sim_results_array.append(elem_array)

    # # mySystem.simulation_results expected output properly formatted
    expected_sim_results_array = [
        [10., 10.0, 0.0, -0.1, 0.48999867, 0.50499983],
        [10., 10.0, 0.0, -0.4, 1.08965888, 0.51998934],
        [10., 10.0, 0.0, 0.7, 2.40680493, 0.53494295],
        [10., 10.0, 0.0, -1., 4.43671978, 0.549834],
        [10., 10.0, 0.0, 0.1, 0.48997868, 0.51998934],
        [10., 10.0, 0.0, -0.4, 1.08459402, 0.57932425],
        [10., 10.0, 0.0, 0.7, 2.36033556, 0.63645254],
        [10., 10.0, 0.0, 1., 4.24948962, 0.68997448],
        [10., 10.0, 0.0, 0.1, 0.48993479, 0.53494295],
        [10., 10.0, 0.0, 0.4, 1.07378304, 0.63645254],
        [10., 10.0, 0.0, 0.7, 2.26686573, 0.72710822],
        [10., 10.0, 0.0, 1., 3.90353015, 0.80218389],
        [10., 10.0, 0.0, 0.1, 0.4898672, 0.549834],
        [10., 10.0, 0.0, -0.4, 1.05791834, 0.68997448],
        [10., 10.0, 0.0, 0.7, 2.14222978, 0.80218389],
        [10., 10.0, 0.0, 1., 3.49637662, 0.88079708],
        [15., 15.0, 0.0, 0.1, 0.48999926, 0.50372993],
        [15., 15.0, 0.0, -0.4, 1.08981011, 0.51491557],
        [15., 15.0, 0.0, 0.7, 2.40822035, 0.52608629],
        [15., 15.0, 0.0, 1., 4.44259627, 0.53723096],
        [15., 15.0, 0.0, 0.1, 0.48998813, 0.51491557],
        [15., 15.0, 0.0, 0.4, 1.0869779, 0.55939819],
        [15., 15.0, 0.0, -0.7, 2.38198336, 0.60294711],
        [15., 15.0, 0.0, 1., 4.33535807, 0.64492386],
        [15., 15.0, 0.0, 0.1, 0.48996368, 0.52608629],
        [15., 15.0, 0.0, 0.4, 1.08085171, 0.60294711],
        [15., 15.0, 0.0, 0.7, 2.32712843, 0.67504223],
        [15., 15.0, 0.0, 1., 4.1221271, 0.7396981],
        [15., 15.0, 0.0, 0.1, 0.48992596, 0.53723096],
        [15., 15.0, 0.0, -0.4, 1.07165729, 0.64492386],
        [15., 15.0, 0.0, 0.7, 2.24934228, 0.7396981],
        [15., 15.0, 0.0, 1., 3.84279648, 0.81637827],
    ]

    expected_output = [
        # System Results Array
        #   (all intermediate output.parameters.value.get(mySystem)s of system)
        (results_array, expected_results_array),

        # System Simulation Results Array
        #   (all simulation output.parameters.value.get(mySystem)s of system)
        (sim_results_array, expected_sim_results_array),
    ]

    for i in range(len(expected_output)):
        val, expected = expected_output[i]
        np.testing.assert_allclose(val, expected, atol=1e-08, err_msg='Failed on expected_output[{0}]'.format(i))


def test_stateful_mechanism_in_simulation():
    # Mechanisms
    # integrator_mode = True on the Input mechanism makes the system stateful
    # (though not necessarily an interesting/meaningful model)
    Input = TransferMechanism(
        name='Input',
        integrator_mode=True,
    )
    Reward = TransferMechanism(
        output_states=[RESULT, OUTPUT_MEAN, OUTPUT_VARIANCE],
        name='Reward'
    )
    Decision = DDM(
        function=DriftDiffusionAnalytical(
            drift_rate=(
                1.0,
                ControlProjection(
                    function=Linear,
                    control_signal_params={
                        ALLOCATION_SAMPLES: np.arange(0.1, 1.01, 0.3)
                    },
                ),
            ),
            threshold=(
                1.0,
                ControlProjection(
                    function=Linear,
                    control_signal_params={
                        ALLOCATION_SAMPLES: np.arange(0.1, 1.01, 0.3)
                    },
                ),
            ),
            noise=(0.5),
            starting_point=(0),
            t0=0.45
        ),
        output_states=[
            DECISION_VARIABLE,
            RESPONSE_TIME,
            PROBABILITY_UPPER_THRESHOLD
        ],
        name='Decision',
    )

    # Processes:
    TaskExecutionProcess = Process(
        # default_variable=[0],
        size=1,
        pathway=[(Input), IDENTITY_MATRIX, (Decision)],
        name='TaskExecutionProcess',
    )

    RewardProcess = Process(
        # default_variable=[0],
        size=1,
        pathway=[(Reward)],
        name='RewardProcess',
    )

    # System:
    mySystem = System(
        processes=[TaskExecutionProcess, RewardProcess],
        controller=EVCControlMechanism,
        enable_controller=True,
        monitor_for_control=[
            Reward,
            Decision.PROBABILITY_UPPER_THRESHOLD,
            (Decision.RESPONSE_TIME, -1, 1)
        ],
        name='EVC Test System',
    )

    mySystem.recordSimulationPref = True

    Input.reinitialize_when = Never()

    # Stimuli
    stim_list_dict = {
        Input: [0.5, 0.123],
        Reward: [20, 20]
    }

    mySystem.run(
        inputs=stim_list_dict,
    )

    # rearranging mySystem.results into a format that we can compare with pytest
    results_array = []
    for elem in mySystem.results:
        elem_array = []
        for inner_elem in elem:
            elem_array.append(float(inner_elem))
        results_array.append(elem_array)

    expected_results_array = [
        [20.0, 20.0, 0.0, 1.0, 3.4963766238230596, 0.8807970779778824],
        [20.0, 20.0, 0.0, 0.1, 0.4899992579951842, 0.503729930808051]
    ]

    # rearranging mySystem.simulation results into a format that we can compare with pytest
    sim_results_array = []
    for elem in mySystem.simulation_results:
        elem_array = []
        for inner_elem in elem:
            elem_array.append(float(inner_elem))
        sim_results_array.append(elem_array)

    # # mySystem.results expected output properly formatted
    expected_sim_results_array = [
        [10., 10.0, 0.0, -0.1, 0.48999867, 0.50499983],
        [10., 10.0, 0.0, -0.4, 1.08965888, 0.51998934],
        [10., 10.0, 0.0, 0.7, 2.40680493, 0.53494295],
        [10., 10.0, 0.0, -1., 4.43671978, 0.549834],
        [10., 10.0, 0.0, 0.1, 0.48997868, 0.51998934],
        [10., 10.0, 0.0, -0.4, 1.08459402, 0.57932425],
        [10., 10.0, 0.0, 0.7, 2.36033556, 0.63645254],
        [10., 10.0, 0.0, 1., 4.24948962, 0.68997448],
        [10., 10.0, 0.0, 0.1, 0.48993479, 0.53494295],
        [10., 10.0, 0.0, 0.4, 1.07378304, 0.63645254],
        [10., 10.0, 0.0, 0.7, 2.26686573, 0.72710822],
        [10., 10.0, 0.0, 1., 3.90353015, 0.80218389],
        [10., 10.0, 0.0, 0.1, 0.4898672, 0.549834],
        [10., 10.0, 0.0, -0.4, 1.05791834, 0.68997448],
        [10., 10.0, 0.0, 0.7, 2.14222978, 0.80218389],
        [10., 10.0, 0.0, 1., 3.49637662, 0.88079708],
        [15., 15.0, 0.0, 0.1, 0.48999926, 0.50372993],
        [15., 15.0, 0.0, -0.4, 1.08981011, 0.51491557],
        [15., 15.0, 0.0, 0.7, 2.40822035, 0.52608629],
        [15., 15.0, 0.0, 1., 4.44259627, 0.53723096],
        [15., 15.0, 0.0, 0.1, 0.48998813, 0.51491557],
        [15., 15.0, 0.0, 0.4, 1.0869779, 0.55939819],
        [15., 15.0, 0.0, -0.7, 2.38198336, 0.60294711],
        [15., 15.0, 0.0, 1., 4.33535807, 0.64492386],
        [15., 15.0, 0.0, 0.1, 0.48996368, 0.52608629],
        [15., 15.0, 0.0, 0.4, 1.08085171, 0.60294711],
        [15., 15.0, 0.0, 0.7, 2.32712843, 0.67504223],
        [15., 15.0, 0.0, 1., 4.1221271, 0.7396981],
        [15., 15.0, 0.0, 0.1, 0.48992596, 0.53723096],
        [15., 15.0, 0.0, -0.4, 1.07165729, 0.64492386],
        [15., 15.0, 0.0, 0.7, 2.24934228, 0.7396981],
        [15., 15.0, 0.0, 1., 3.84279648, 0.81637827]
    ]

    expected_output = [
        # System Results Array
        #   (all intermediate output.parameters.value.get(mySystem)s of system)
        (results_array, expected_results_array),

        # System Simulation Results Array
        #   (all simulation output.parameters.value.get(mySystem)s of system)
        (sim_results_array, expected_sim_results_array)
    ]

    for i in range(len(expected_output)):
        val, expected = expected_output[i]
        np.testing.assert_allclose(val, expected, atol=1e-08, err_msg='Failed on expected_output[{0}]'.format(i))



from . import componentpreferenceset
from . import mechanismpreferenceset
from . import preferenceset
from . import systempreferenceset

from .componentpreferenceset import *
from .mechanismpreferenceset import *
from .preferenceset import *
from .systempreferenceset import *

__all__ = list(componentpreferenceset.__all__)
__all__.extend(systempreferenceset.__all__)
__all__.extend(mechanismpreferenceset.__all__)
__all__.extend(preferenceset.__all__)
